#!/usr/bin/env python
from intervaltree import Interval, IntervalTree
import networkx as nx
import argparse
import os
import sys
import time

#reveal imports
import schemes
import transform
import transformold
import plot
import utils
import merge
import subgraph
import comp
import extract
import convert
import rem
import refine
import bubbles
import matches
import chain
import stats
import split
import align
import unzip
import chop
import annotate


import logging
#add custom loglevel TRACE
logging.TRACE = 1
logging.addLevelName(logging.TRACE, "TRACE")
logging.logThreads = 0
logging.Logger.trace = lambda inst, msg, *args, **kwargs: inst.log(logging.TRACE, msg, *args, **kwargs)
logging.trace = lambda msg, *args, **kwargs: logging.log(logging.TRACE, msg, *args, **kwargs)

def main():
    desc="""
    Type 'reveal <positional_argument> --help' for help on a specific subcommand.\n
    Reveal constructs population reference graphs by aligning multiple whole 
    genomes using recursive exact matching.
    http://www.biorxiv.org/content/early/2015/07/17/022715.
    """
    
    parser = argparse.ArgumentParser(prog="reveal", usage="reveal -h for usage", description=desc, formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    global_parser = argparse.ArgumentParser(add_help=False) #parser for arguments that apply to all subcommands
    global_parser.add_argument("-l", "--log-level", type=int, dest="loglevel", default=20, help="Log level: 1=trace 10=debug 20=info 30=warn 40=error 50=fatal.")
    global_parser.add_argument("--64", dest="sa64", default=False, action="store_true", help="Use 64bit suffix array in the index.")
    
    subparsers = parser.add_subparsers()
    
    parser_aln = subparsers.add_parser('align',prog="reveal align", description="Output a bash-script that decribes the various steps that need to be performed to construct a graph-based multi-genome for the input (draft-)genomes.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_aln.add_argument('reference', nargs=1, help='Reference assembly to which draft-asemblies should be organized (this may be a draft assembly as well, just used for organizing structural rearrangements).')
    parser_aln.add_argument('inputfiles', nargs='+', help='(Multi-)Fasta files that specify the draft-assemblies that are to be aligned (.fasta).')
    parser_aln.add_argument("-o", "--output", dest="output", default="prg", help="Prefix for the filename of the resulting graph.")
    parser_aln.add_argument("-m", dest="m", default=20, type=int, help="Min length of an anchor to constrain the alignment.")
    parser_aln.add_argument("-n", dest="n", default=None, type=int, help="Number of genomes for anchor to constrain the alignment.")
    parser_aln.add_argument("--chunksize", dest="chunksize", type=int, default=2, help="If order is sequential, use this many genomes per alignment.")
    parser_aln.add_argument("--nproc", dest="nproc", default=1, type=int, help="Number of processes to use for the individual steps.")
    parser_aln.add_argument("--order", dest="order", default="simultaneous", choices=["simultaneous","sequential","tree"], help="Order in which graph is constructed.")
    parser_aln.add_argument("--norefine", dest="refine", action="store_false", default=True, help="Do not use consistency based msa, just produce the anchor graph.")
    parser_aln.add_argument("--minconf", dest="minconf", type=float, default=90, choices=range(0,100), metavar="[0-99]", help="Use cutoff on confidence values during refinement of the graph.")
    parser_aln.add_argument("--nounzip", dest="unzip", action="store_false", default=True, help="Do not unzip bubbles before refining.")
    parser_aln.add_argument("--notransform", dest="transform", action="store_false", default=True, help="Do not use transform to account for structural events or draft assemblies. Assume colinear alignment of complete genomes.")
    parser_aln.add_argument("--novariants", dest="variants", action="store_false", default=True, help="Do not output bubbles as variants.")
    parser_aln.set_defaults(func=align.align)

    parser_rem = subparsers.add_parser('rem',prog="reveal rem", description="Use recursive exact matching to obtain a graph from multiple input genomes or other graphs.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_rem.add_argument('inputfiles', nargs='+', help='Fasta or gfa files specifying either assembly/alignment graphs (.gfa) or sequences (.fasta).')
    parser_rem.add_argument("-o", "--output", dest="output", help="Prefix of the alignment graph.")
    parser_rem.add_argument("-t", "--threads", dest="threads", type=int, default=0, help = "The number of threads to use for the alignment.")
    parser_rem.add_argument("-m", dest="minlength", type=int, default=20, help="Min length of an exact match.")
    parser_rem.add_argument("-p", dest="pcutoff", type=float, default=1e-08, help="Use this significance threshold for exact matches, when -m=0")
    parser_rem.add_argument("-n", dest="minn", type=int, default=2, help="Only align graph on exact matches that occur in at least this many samples (if not set, equal to total number of genomes in the (sub)index).")
    parser_rem.add_argument("--gcmodel", dest="gcmodel", choices=["sumofpairs","star-avg","star-med"], default="sumofpairs", help="Which gap-cost model to use.")
    parser_rem.add_argument("--wp", dest="wpen", type=int, default=1, help="Weight of penalty during chaining.")
    parser_rem.add_argument("--ws", dest="wscore", type=int, default=1, help="Weight of score during chaining.")
    parser_rem.add_argument("--seedsize", dest="seedsize", type=int, default=10000, help="Skip recursion for chained mums larger than this size (when 0 don't seed).")
    parser_rem.add_argument("--maxmums", dest="maxmums", type=int, default=1000, help="Number of largest MUMs to consider for chaining per iteration (when 0 use all).")
    parser_rem.add_argument("--plot", dest="mumplot", action="store_true", default=False, help="Save a mumplot for the actual aligned chain of anchors (depends on matplotlib).")
    parser_rem.add_argument("-i", dest="interactive", action="store_true", default=False, help="Show an interactive visualisation of the mumplot (depends on matplotlib).")
    parser_rem.add_argument("--sa", dest="sa", default="", help="Specify a preconstructed suffix array to decouple suffix array construction.")
    parser_rem.add_argument("--lcp", dest="lcp", default="", help="Specify a preconstructed lcp array to decouple lcp array construction.")
    parser_rem.add_argument("--cache", dest="cache", default=False, action="store_true", help="When specified, it caches the suffix and lcp array to disk after construction.")
    parser_rem.add_argument("-g", dest="minsamples", type=int, default=1, help="Only index nodes that occur in this many samples or more.")
    parser_rem.add_argument("-x", dest="maxsamples", type=int, default=None, help="Only align nodes that have maximally this many samples.")
    parser_rem.add_argument("-r", dest="reference", type=str, default=None, help="Name of the sequence that should be used as a coordinate system or reference.")
    parser_rem.add_argument("-s", dest="targetsample", type=str, default=None, help="Only align nodes in which this sample occurs.")    
    parser_rem.add_argument("--gml", dest="gml", action="store_true", default=False, help="Produce a gml graph instead gfa.")
    parser_rem.add_argument("--gml-max", dest="hwm", default=4000, help="Max number of nodes per graph in gml output.")
    parser_rem.add_argument("--noupper", dest="toupper", action="store_false", default=True, help="Do not force uppercase for sequence, in case we want to prevent matches in repeatmasked sequence.")
    parser_rem.add_argument("--maxbubblesize", dest="maxsize", type=int, default=None, help="Apply recursion until largest allele within a bubble is smaller than this size.")
    parser_rem.add_argument("--nocontigs", dest="contigs", default=True, action="store_false", help="Don't treat multi-fasta files as contigs, use every sequence as a target.")
    parser_rem.add_argument("--notrim", dest="trim", default=True, action="store_false", help="Don't trim overlap between MUMs, thus more greedy positioning of indels.")
    parser_rem.set_defaults(func=rem.align_cmd)

    parser_unzip = subparsers.add_parser('unzip',prog="reveal unzip", description="Opens up bubbles to account for uncertainty of indel placement and edge-wander. Specify --source and --sink to unzip a specific bubble.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_unzip.add_argument("graph", nargs=1, help='Graph in gfa format for which bubbles should be unzipped.')
    parser_unzip.add_argument("-u", dest="minunzip", type=int, default=0, help="Try to unzip all bubbles at least this many bases.")
    parser_unzip.add_argument("-o", "--output", dest="output", default=None, help="Prefix for the filename of the resulting graph.")
    parser_unzip.add_argument("--mindiff", dest="mindiff", default=1, type=int, help="Only unzip bubbles where the difference between the min- and max-allele size is larger than this many bp, by default 1, so don't unzip SNPs.")
    parser_unzip.add_argument("--maxdiff", dest="maxdiff", default=10000, type=int, help="Only unzip bubbles where the difference between the min- and max-allele size is smaller than this many bp.")
    parser_unzip.add_argument("--source", dest="source", type=int, default=None, help="Source for specific bubble.")
    parser_unzip.add_argument("--sink", dest="sink", type=int, default=None, help="Sink for specific bubble.")
    parser_unzip.set_defaults(func=unzip.unzip)

    parser_chop = subparsers.add_parser('chop',prog="reveal chop", description="Uses the chop algorithm to introduce overlap (of length k-1) onto the edges of the graph such that reads of length k can be mapped on to the graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_chop.add_argument("graph", nargs=1, help='Graph in gfa format which has to be chopped.')
    parser_chop.add_argument("-k", dest="k", type=int, default=100, help="Max length of the reads that need to be mapped to the graph.")
    parser_chop.add_argument("-o", "--output", dest="output", default=None, help="Prefix for the resulting overlap graph and fasta file.")
    parser_chop.add_argument("--noextend", dest="extend", default=True, action="store_false", help="Do not add prefix/suffix, just apply duplicate and contract.")
    parser_chop.add_argument("--fasta", dest="fasta", default=False, action="store_true", help="Write node sequence to a fasta file, for read mapping.")
    parser_chop.add_argument("--width", dest="lw", type=int, default=100, help="Line width for fasta output.")
    parser_chop.add_argument("--check", dest="check", default=False, action="store_true", help="Check if all k-length substrings of the input haplotype are covered in the flat representation.")
    parser_chop.set_defaults(func=chop.chop_cmd)

    parser_refine = subparsers.add_parser('refine', prog="reveal refine", description="Refine bubbles in the graph by multiple sequence alignment.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_refine.add_argument("graph", nargs=1, help='Graph in gfa format for which bubbles should be realigned.')
    parser_refine.add_argument("source", nargs='?', default=None, type=int, help='Source node.')
    parser_refine.add_argument("sink", nargs='?', default=None, type=int, help='Sink node.')
    parser_refine.add_argument("--nproc", dest="nproc", default=1, type=int, help="Use multiprocessing to realign bubbles.")
    parser_refine.add_argument("--chunksize", dest="chunksize", type=int, default=10, help="Process in parallel refined bubbles in chunks of this size.")
    parser_refine.add_argument("--method", dest="method", choices=["reveal_probcons","reveal_rem","muscle","probcons","msaprobs","pecan"], default="reveal_probcons", help="Use external multiple sequence aligner for the alignment of bubbles (expects methods to be accessible through the $PATH variable, reveal_* methods use internal memory, other methods uese external memory")
    parser_refine.add_argument("--params", dest="parameters", default="", help="Add this value when calling the external methods.")
    parser_refine.add_argument("-o", dest="outfile", type=str, default=None, help="File to which realigned graph is to be written.")
    parser_refine.add_argument("--all", action="store_true", dest="all", default=False, help="Trigger realignment for all bubbles.")
    parser_refine.add_argument("--complex", action="store_true", dest="complex", default=False, help="Trigger realignment for all complex bubbles.")
    parser_refine.add_argument("--nogaps", action="store_true", dest="nogaps", default=False, help="Skip realignment for bubbles that span gaps.")
    parser_refine.add_argument("--simple", action="store_true", dest="simple", default=False, help="Trigger realignment for all simple bubbles.")
    parser_refine.add_argument("--minsize", dest="minsize", type=int, default=0, help="Only realign bubbles if the smallest allele contains at least this many bases.")
    parser_refine.add_argument("--maxsize", dest="maxsize", type=int, default=10000, help="Only realign bubbles if the largest allele contains less than this many bases.")
    # parser_refine.add_argument("--minmaxsize", dest="minmaxsize", type=int, default=2, help="Only realign bubbles if the largest allele contains more than this many bases.")
    parser_refine.add_argument("--mindiff", dest="mindiff", default=1, type=int, help="Only refine variants where the difference between the min- and max-allele size is larger than this many bp.")
    parser_refine.add_argument("--maxdiff", dest="maxdiff", default=None, type=int, help="Only refine variants where the difference between the min- and max-allele size is smaller than this many bp.")
    parser_refine.add_argument("--maxcumsize", dest="maxcumsize", type=int, default=None, help="Maximum length of the cumulative sum of all paths that run through the bubble.")
    parser_refine.add_argument("--mincumsize", dest="mincumsize", type=int, default=0, help="Minimum length of the cumulative sum of all paths that run through the bubble.")
    parser_refine.add_argument("--minconf", dest="minconf", type=float, default=0, choices=range(0,101), metavar="[0-100]", help="Use cutoff on confidence values from the MSA in graph construction ().")
    parser_refine.add_argument("--uniqueonly", dest="uniqueonly", default=False, action="store_true", help="Only consider unique haplotypes for multiple sequence alignment.")
    parser_refine.add_argument("-c","--consistency", dest="constrans", type=int, default=2, help="Number of consistency transformations to apply before alignment (only applies to reveal_probcons).")
    parser_refine.add_argument("-r","--iterative-refinement", dest="nrefinements", type=int, default=100, help="Number of iterative refinements to apply after alignment (only applies to reveal_probcons).")
    parser_refine.add_argument("--no-gap-consistency", dest="consgap", action="store_false", default=True, help="Don't consider gaps in consistency transform (only applies to reveal_probcons).")
    parser_refine.add_argument("-m", dest="minlength", type=int, default=20, help="Min length of an exact match (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("-n", dest="minn", type=int, default=2, help="Only align graph on exact matches that occur in at least this many samples (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("--gcmodel", dest="gcmodel", choices=["sumofpairs","star-avg","star-med"], default="sumofpairs", help="Which gap-cost model to use for multi-alignment (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("--wp", dest="wpen", type=int, default=1, help="Multiply penalty for a MUM by this number in scoring scheme (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("--ws", dest="wscore", type=int, default=1, help="Multiply length of MUM by this number in scoring scheme (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("--seedsize", dest="seedsize", type=int, default=10000, help="Skip recursion for chained mums larger than this size (when 0 don't seed) (only applies when method is 'reveal_rem').")
    parser_refine.add_argument("--maxmums", dest="maxmums", type=int, default=1000, help="Number of largest MUMs to consider for chaining (when 0 use all) (only applies when method is 'reveal_rem').")
    parser_refine.set_defaults(func=refine.refine_bubble_cmd)
    parser_realign = subparsers.add_parser('realign', prog="reveal realign", parents=[parser_refine], add_help=False)

    parser_extract = subparsers.add_parser('extract', prog="reveal extract", description="Extract the input sequence from a graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_extract.add_argument('graph', nargs=1, help='gfa file specifying the graph from which the genome should be extracted.')
    parser_extract.add_argument('input', nargs='*', help='Name of the sample or path to be extracted from the graph.')
    parser_extract.add_argument("-t", dest="type", default="pathname", choices=["pathname","path"], help="Type of input, either pathname or comma-seperated sequence of node-ids.")
    parser_extract.add_argument("--width", dest="width", type=int, default=100 , help='Line width for fasta output.')
    parser_extract.add_argument("--all", dest="all", default=False, action="store_true", help="Extract all paths from the graph and output as fasta.")
    parser_extract.add_argument("--nocycles",  action="store_true", dest="nocycles", default=False, help="Parse only the directed acyclic layout of the graph, so ignore strucural rearrangements (cycles) in the graph.")
    parser_extract.set_defaults(func=extract.extract_cmd)
    
    parser_plot = subparsers.add_parser('plot', prog="reveal plot", description="Generate a mumplot that shows all mums between two fasta files.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_plot.add_argument('fastas', nargs='*', help='Two fasta files for which a mumplot should be generated.')
    parser_plot.add_argument("-m", dest="minlength", type=int, default=20, help="Minimum length of exact matches to vizualize.")
    parser_plot.add_argument("-i", dest="interactive", action="store_true", default=False, help="Wheter to produce interactive plots which allow zooming on the dotplot.")
    parser_plot.add_argument("--endpoints", dest="endpoints", action="store_true", default=False, help="Mark mum start/end points.")
    parser_plot.add_argument("--norc", dest="rc", action="store_false", default=True, help="Don't draw reverse complement matches.")
    parser_plot.add_argument("--maxmums", dest="maxmums", type=int, default=10000, help="Cap the number of MUMs to draw.")
    parser_plot.add_argument("--nogaps", dest="showgaps", action="store_false", default=True, help="Don't mark gapped sequence.")
    parser_plot.add_argument("--extension", dest="extension", choices=["png","pdf","ps","eps","svg"], default="png", help="How to save the plot.")
    parser_plot.add_argument("-r","--xr", dest="xregion", default=None, help="Highlight and zoom on intervals (encoded as \"<start1>-<end1>,<start2>-<end2>\" etc.) with respect to x-axis (first sequence).")
    parser_plot.add_argument("--yr", dest="yregion", default=None, help="Highlight and zoom on intervals (encoded as \"<start1>-<end1>,<start2>-<end2>\" etc.) with respect to y-axis (second sequence).")
    parser_plot.add_argument("--flanksize", dest="flanksize", default=None, help="In case of (a) specified region include this many bases of flanking sequence (encode as \"<flanksize_region1>,<flanksize_region2>\" etc.).")
    parser_plot.set_defaults(func=plot.plot)
    
    parser_gplot = subparsers.add_parser('gplot', prog="reveal gplot", description="Generate a plot that represents the alignment of two samples in a graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_gplot.add_argument('graph', help='A graph representing the genomes of two or more samples.')
    parser_gplot.add_argument("-m", dest="minlength", type=int, default=1, help="Minimum length of exact matches to vizualize.")
    parser_gplot.add_argument("-x", default=None, help='Name of sample 1 (x-axis), when graph contains more than two samples, assignment is random.')
    parser_gplot.add_argument("-y", default=None, help='Name of sample 2 (y-axis), when graph contains more than two samples, assignment is random.')
    parser_gplot.add_argument("-i", dest="interactive", action="store_true", default=False, help="Wheter to produce interactive plots which allow zooming on the dotplot.")
    parser_gplot.add_argument("-r", dest="region", default=None, help="Highlight interval (as \"<start>:<end>\") with respect to x-axis (first sequence).")
    parser_gplot.set_defaults(func=plot.gplot)
    
    parser_comp = subparsers.add_parser('comp', prog="reveal comp", description="Reverse complement the graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_comp.add_argument('graph', nargs=1, help='The graph to be reverse complemented.')
    parser_comp.set_defaults(func=comp.comp_cmd)
    

    # #the old transform method
    # parser_transform = subparsers.add_parser('transform', prog="reveal transform", description="Transform a draft assembly into a graph that encodes the structural order of assembled segments with respect to a finished reference assembly.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    # parser_transform.add_argument('reference', help='(Multi-)fasta reference sequence.')
    # parser_transform.add_argument('contigs', help='(Multi-)fasta draft assembly that contains contigs that are to be oriented and ordered with respect to the reference.')
    # parser_transform.add_argument("-o", "--output", dest="output", help="Prefix of fasta file for the \'finished\' genome.")
    # parser_transform.add_argument("-m", dest="minlength", type=int, default=20, help="Min length of maximal exact matches for considering (if set to 0, try to extract all MUMs).")
    # parser_transform.add_argument("-i", dest="interactive", action="store_true", default=False, help="Output interactive plot.")
    # parser_transform.add_argument("--nproc", dest="nproc", default=1, type=int, help="Use multiprocessing to do MUM extraction (max: 2 proc) and mapping (max: number of contigs) in parallel (increases mem usage!).")
    # parser_transform.add_argument("--gcmodel", dest="gcmodel", choices=["sumofpairs","star-avg","star-med"], default="sumofpairs", help="Which gap-cost model to use for multi-alignment.")
    # parser_transform.add_argument("--plot", dest="plot", action="store_true", default=False, help="Output mumplots for the \'finished\' chromosomes (depends on matplotlib).")
    # parser_transform.add_argument("--graph", dest="outputtype", choices=["graph","fasta"], default="graph", help="Output a graph or fasta representation of the transformed genome.")
    # parser_transform.add_argument("--structonly", dest="allcontigs", action="store_false", default=True, help="Only output paths for contigs that contain structural rearrangements.")
    # # parser_transform.add_argument("--filter", dest="filtermums", action="store_true", default=False, help="Reduce search space by filtering exact matches.")
    # parser_transform.add_argument("--plotall", dest="plotall", action="store_true", default=False, help="Plot all matches, instead of only the chained matches.")
    # parser_transform.add_argument("--split", dest="split", action="store_true", default=False, help="Split the \'finished\' genome by chromosome.")
    # parser_transform.add_argument("--order", dest="order", default="chains", choices=["contigs","chains"], help="Determine the order for either contigs or chains.")
    # parser_transform.add_argument("--mineventsize", dest="mineventsize", type=int, default=200, help="Maximal distance between clusters/mums for chaining.")
    # parser_transform.add_argument("--minchainsum", dest="minchainsum", type=int, default=10000, help="Minimal sum of the length of the MUMs in a chain before its considered.")
    # parser_transform.add_argument("--maxmums", dest="maxmums", type=int, default=0, help="Max number of MUMs to consider for chaining (when 0 use all).")
    # parser_transform.add_argument("--cutn", dest="cutn", type=int, default=1000, help="Cut contigs at N-stretches longer than this value, to force re-estimation of gapsizes (set to 0, to switch off).")
    # parser_transform.add_argument("--fixedgapsize", dest="fixedsize", action="store_true", default=False, help="Do not estimate gapsize based on reference, instead use fixed gapsizes of length that can be set with \'gapsize\'.")
    # parser_transform.add_argument("--gapsize", dest="gapsize", type=int, default=100, help="Use this number of N's between adjacent (only in case of fixedgapsizes) or  partially overlapping contigs.")
    # parser_transform.add_argument("--maxdist", dest="maxdist", type=int, default=90, help="Max space between adjacent MUMs in a cluster.")
    # parser_transform.add_argument("--mincluster", dest="mincluster", type=int, default=65, help="Max space between adjacent MUMs in a cluster.")
    # parser_transform.add_argument("--extiter", dest="extiter", type=int, default=3, help="Number of extension iterations using locally unique MUMs.")
    # parser_transform.add_argument("--maxextend", dest="maxextend", type=int, default=1000, help="Size of the region to try to inspect for locally unique MUMs.")
    # parser_transform.add_argument("--ml", dest="minlocallength", type=int, default=20, help="Min size of locally unique mums.")
    # parser_transform.add_argument("--unmapped", dest="outputunmapped", action="store_true", default=False, help="Output a unmappable sequence to a separate fasta file.")
    # parser_transform.set_defaults(func=transform.transform)

    
    parser_transform = subparsers.add_parser('transform', prog="reveal transform", description="Transform a draft assembly into a graph that encodes the structural order of assembled segments with respect to a finished reference assembly.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])

    parser_transform.add_argument('reference', nargs=1, help='(Multi-)fasta reference sequence.')
    parser_transform.add_argument('contigs', nargs='+', help='(Multi-)fasta draft assembly that contains contigs that are to be oriented and ordered with respect to the reference.')

    parser_transform.add_argument("--cutn", dest="cutn", type=int, default=0, help="Cut contigs at N-stretches longer than this value (default is 0, off).")
    parser_transform.add_argument("-o", "--output", dest="output", help="Prefix of gfa file for the \'transformed\' genome.")
    parser_transform.add_argument("-c", dest="minctglength", type=int, default=10000, help="Skip transform for contigs short than this length.")
    parser_transform.add_argument("-m", dest="minlength", type=int, default=20, help="Min length of maximal exact matches for considering (if set to 0, try to extract all MUMs).")
    parser_transform.add_argument("-i", dest="interactive", action="store_true", default=False, help="Output interactive plot.")
    parser_transform.add_argument("--plot", dest="plot", action="store_true", default=False, help="Output mumplots for the \'finished\' chromosomes (depends on matplotlib).")

    parser_transform.add_argument("--rc", dest="rearrangecost", default=10000, type=int, help="Cost for chaining translocated segments.")
    parser_transform.add_argument("--ic", dest="inversioncost", default=5, type=int, help="Cost for chaining inverted segments.")
    
    parser_transform.add_argument("--alfa", dest="alfa", default=2, type=float, help="Weight for match (+).")
    parser_transform.add_argument("--lambda", dest="_lambda", default=3, type=float, help="Weight for indel penalty (-).")
    parser_transform.add_argument("--eps", dest="eps", default=2, type=float, help="Weight for substitution penalty (-).")
    parser_transform.add_argument("--gapopen", dest="gapopen", default=1, type=float, help="Fixed penalty for adding an achor to the chain (-).")
    
    parser_transform.add_argument("--nocluster", dest="cluster", action="store_false", default=True, help="Don't cluster MUMs by diagonals.")
    parser_transform.add_argument("--maxdist", dest="maxdist", type=int, default=30, help="Max space between adjacent MUMs (on the same diagonal) in a cluster.")
    parser_transform.add_argument("--mincluster", dest="mincluster", type=int, default=50, help="Minimal size (sum of mums) of a cluster.")
    parser_transform.add_argument("--minchainsum", dest="minchainsum", type=int, default=50, help="Minimal sum of the length of the MUMs in a chain before its considered.")

    parser_transform.add_argument("--noopt", dest="optimise", action="store_false", default=True, help="Don't perform naive optimisation of the glocal chain.")
    parser_transform.add_argument("--heap", dest="useheap", action="store_true", default=False, help="Use a priority queue to compute an optimal chain.")
    parser_transform.add_argument("--lastn", dest="lastn", type=int, default=50, help="Backtrack at least this many anchors while chaining fragments.")
    parser_transform.add_argument("--lastbp", dest="lastbp", type=int, default=20000, help="Backtrack at least this many bp while chaining fragments.")

    parser_transform.add_argument("--greedy", dest="greedy", action="store_true", default=False, help="Assign overlap between anchors in a greedy manner. Large anchors become larger.")
    parser_transform.add_argument("--outputbed", dest="outputbed", action="store_true", default=True, help="Produce a bed file that stores the rearrangement breakpoints on the reference assembly.")

    parser_transform.set_defaults(func=transform.transform_cmd)







    parser_finish = subparsers.add_parser('finish', prog="reveal finish", description="Order and orient the contigs of a draft assembly with respect to a finished reference assembly (essentially the same as transform but with different default parameters).", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_finish.add_argument('reference', help='(Multi-)fasta reference sequence.')
    parser_finish.add_argument('contigs', help='(Multi-)fasta draft assembly that contains contigs that are to be oriented and ordered with respect to the reference.')
    parser_finish.add_argument("-o", "--output", dest="output", help="Prefix of fasta file for the \'finished\' genome.")
    parser_finish.add_argument("-m", dest="minlength", type=int, default=15, help="Min length of maximal exact matches for considering (if not set, use the set of largest MUMs for which the genome wide coverage is below 1).")
    parser_finish.add_argument("-i", dest="interactive", action="store_true", default=False, help="Output interactive plot.")
    
    parser_finish.add_argument("--nproc", dest="nproc", default=1, type=int, help="Use multiprocessing to do MUM extraction (max: 2 proc) and mapping (max: number of contigs) in parallel (increases mem usage!).")
    parser_finish.add_argument("--gcmodel", dest="gcmodel", choices=["sumofpairs","star-avg","star-med"], default="sumofpairs", help="Which gap-cost model to use for multi-alignment.")
    parser_finish.add_argument("--plot", dest="plot", action="store_true", default=False, help="Output mumplots for the \'finished\' chromosomes (depends on matplotlib).")
    parser_finish.add_argument("--graph", dest="outputtype", choices=["graph","fasta"], default="fasta", help="Output a graph or fasta representation of the transformed genome.")
    parser_finish.add_argument("--allcontigs", dest="allcontigs", action="store_true", default=True, help="Output all contigs as separate paths through the graph, otherwise only report contigs that contain structural variants.")
    parser_finish.add_argument("--filter", dest="filtermums", action="store_true", default=False, help="Reduce search space by filtering exact matches.")
    parser_finish.add_argument("--plotall", dest="plotall", action="store_true", default=False, help="Plot all matches, instead of only the chained matches.")
    parser_finish.add_argument("--split", dest="split", action="store_true", default=False, help="Split the \'finished\' genome by chromosome.")
    parser_finish.add_argument("--order", dest="order", default="contigs", choices=["contigs","chains"], help="Determine the order for either contigs or chains.")
    parser_finish.add_argument("--mineventsize", dest="mineventsize", type=int, default=1500, help="Maximal distance between clusters/mums for chaining.")
    parser_finish.add_argument("--minchainsum", dest="minchainsum", type=int, default=1000, help="Minimal sum of the length of the MUMs in a chain before its considered.")
    parser_finish.add_argument("--maxmums", dest="maxmums", type=int, default=0, help="Max number of MUMs to consider for chaining (when 0 use all).")
    parser_finish.add_argument("--cutn", dest="cutn", type=int, default=1000, help="Cut contigs at N-stretches longer than this value, to force re-estimation of gapsizes (set to 0, to switch off).")
    parser_finish.add_argument("--fixedgapsize", dest="fixedsize", action="store_true", default=False, help="Do not estimate gapsize based on reference, instead use fixed gapsizes of length that can be set with \'gapsize\'.")
    parser_finish.add_argument("--gapsize", dest="gapsize", type=int, default=100, help="Use this number of N's between adjacent (only in case of fixedgapsizes) or  partially overlapping contigs.")
    
    parser_finish.add_argument("--maxdist", dest="maxdist", type=int, default=90, help="Max space between adjacent MUMs in a cluster.")
    parser_finish.add_argument("--mincluster", dest="mincluster", type=int, default=20, help="Max space between adjacent MUMs in a cluster.")
    
    parser_finish.add_argument("--extiter", dest="extiter", type=int, default=3, help="Number of iterations of alignment extension.")
    parser_finish.add_argument("--maxextend", dest="maxextend", type=int, default=200, help="Size of the region to try to inspect for locally unique MUMs.")
    parser_finish.add_argument("--ml", dest="minlocallength", type=int, default=20, help="Min size of locally unique mums.")
    
    parser_finish.add_argument("--nounmapped", dest="outputunmapped", action="store_false", default=True, help="Do not output unmappable sequence to a separate fasta file.")
    parser_finish.set_defaults(func=transformold.transform)




    parser_convert = subparsers.add_parser('convert', prog="reveal convert", description="Convert gfa graph to gml.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_convert.add_argument('graphs', nargs='*', help='The gfa graph to convert to gml.')
    parser_convert.add_argument("-n", dest="minsamples", type=int, default=1, help="Only align nodes that occcur in this many samples.")
    parser_convert.add_argument("-x", dest="maxsamples", type=int, default=None, help="Only align nodes that have maximally this many samples.")
    parser_convert.add_argument("-s", dest="targetsample", type=str, default=None, help="Only align nodes in which this sample occurs.")
    parser_convert.add_argument("--gml-max", dest="hwm", default=4000, type=int, help="Max number of nodes per graph in gml output.")
    # parser_convert.add_argument("--gfa",  action="store_true", dest="gfa", default=False, help="Rewrite gfa file.")
    parser_convert.add_argument("--partition",  action="store_true", dest="partition", default=False, help="Output graph as multiple subgraphs if possible.")
    parser_convert.add_argument("--nocycles",  action="store_true", dest="nocycles", default=False, help="Do not allow rearrangements (cycles) in graph.")
    parser_convert.add_argument("--to", dest="type", default="gml", choices=['gml','gfa','maf'], help="Filetype to convert to.")
    parser_convert.add_argument("--aligned", dest="aligned", default=False, action="store_true", help="Whether multi fasta file is aligned.")
    parser_convert.set_defaults(func=convert.convert)
    
    parser_subgraph = subparsers.add_parser('subgraph', prog="reveal subgraph", description="Extract subgraph from gfa by specified node ids.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_subgraph.add_argument('inputfiles', nargs='*', help="The gfa graph followed by a subgraph definition. Either comma-separated node ids (e.g. \"node1,node2,node3,...\"), topological range (all nodes between e.g. \"<node1>-<node2>\"), interval-based (e.g. \"chr4:34000-35000\").")
    parser_subgraph.add_argument("-o", dest="outfile", type=str, default="subgraph", help="Prefix of the file to which subgraph will be written.")
    parser_subgraph.add_argument("--gml", dest="gml", action="store_true", default=False, help="Produce a gml graph instead of gfa.")
    parser_subgraph.set_defaults(func=subgraph.subgraph)
    
    parser_bubbles = subparsers.add_parser('bubbles', prog="reveal bubbles", description="Extract all bubbles from the graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_bubbles.add_argument("graph", nargs=1, help='Graph in gfa format from which bubbles are to be extracted.')
    parser_bubbles.add_argument("-e", dest="exportcomplex", action="store_true", default=False, help="Output complex bubble structures in a separate gfa file.")
    parser_bubbles.add_argument("-s", dest="separate", action="store_true", default=False, help="Write a seperate gfa file for each complex bubble structure.")
    parser_bubbles.add_argument("--gml", dest="gml", action="store_true", default=False, help="Output gml instead of gfa.")
    parser_bubbles.set_defaults(func=bubbles.bubbles_cmd)
    
    parser_variants = subparsers.add_parser('variants', prog="reveal variants", description="Extract variant calls from the graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_variants.add_argument("graph", nargs=1, help='Graph in gfa format from which bubbles are to be extracted.')
    parser_variants.add_argument("-r", dest="reference", type=str, default=None, help="Name of the sequence that, if possible, should be used as a coordinate system or reference.")
    parser_variants.add_argument("--fasta", dest="fastaout", action="store_true", default=False, help="Output variant sequence in a fasta format.")
    parser_variants.add_argument("--bed", dest="bedout", action="store_true", default=False, help="Output position of variants in bed format.")
    parser_variants.add_argument("--vcf", dest="vcfout", action="store_true", default=False, help="Output variants in vcf format.")
    parser_variants.add_argument("--split", dest="split", action="store_true", default=False, help="Output a multi-fasta file per variant.")
    parser_variants.add_argument("--minsize", dest="minsize", default=0, type=int, help="Only output variants where max-allele size is larger than this many bp.")
    parser_variants.add_argument("--mindiff", dest="mindiff", default=0, type=int, help="Only output variants where the difference between the min- and max-allele size is larger than this many bp.")
    parser_variants.add_argument("--maxdiff", dest="maxdiff", default=None, type=int, help="Only output variants where the difference between the min- and max-allele size is smaller than this many bp.")
    parser_variants.add_argument("--minflank", dest="minflank", default=0, type=int, help="Only output variants with an exact matching flanking sequence of at least this length.")
    parser_variants.add_argument("--type", dest="type", default="all", choices=["all","snv","indel","multi-allelic","region","complex","undefined"], help="Only output variants of this type.")
    parser_variants.add_argument("--nogaps", dest="nogaps", default=False, action="store_true", help="Don't output variants that are caused by gaps (contain the N character).")
    parser_variants.add_argument("--refonly", dest="refonly", default=False, action="store_true", help="Don't output variants that are not positionable on the specified reference.")
    parser_variants.set_defaults(func=bubbles.variants_cmd)
    
    parser_rearrangements = subparsers.add_parser('rearrangements', prog="reveal rearrangements", description="Report on edges in the graph that describe rearrangements.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_rearrangements.add_argument("-r", dest="reference", type=str, default=None, help="Name of the sequence that, should be used as a coordinate system or reference.")
    parser_rearrangements.add_argument("graph", nargs=1, help='Graph in gfa format for rearrangement edges are reported.')
    parser_rearrangements.set_defaults(func=bubbles.rearrangements_cmd)

    parser_annotate = subparsers.add_parser('annotate', prog="reveal annotate", description="Add annotations to variants in a vcf file using trf and repeatmasker.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_annotate.add_argument("--species", dest="species", type=str, default="human", help="Which \'species\' flag to pass on to repeatmasker.")
    parser_annotate.add_argument("--nproc", dest="repmproc", type=int, default=1, help="How many processes repeatmasker should use for annotation (-pa flag).")
    parser_annotate.add_argument("--mindiff", dest="mindiff", default=50, type=int, help="Only annotate variants where the difference between the min- and max-allele size is larger or equal to this many bp.")
    parser_annotate.add_argument("--maxdiff", dest="maxdiff", default=100000, type=int, help="Only annotate variants where the difference between the min- and max-allele size is smaller than this many bp.")
    parser_annotate.add_argument("--notrf", dest="trf", default=True, action="store_false", help="Skip the tandem repeat finder for annotation.")
    parser_annotate.add_argument("--norepm", dest="repm", default=True, action="store_false", help="Skip the RepeatMasker for annotation.")
    parser_annotate.add_argument("vcffile", help='Variants from a graph in the vcf file format')
    parser_annotate.set_defaults(func=annotate.annotate)

    parser_merge = subparsers.add_parser('merge', prog="reveal merge", description="Combine multiple gfa graphs into a single gfa graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_merge.add_argument("graphs", nargs='*', help='Graphs in gfa format that should be merged.')
    parser_merge.add_argument("-o", dest="outprefix", type=str, default=None, help="Prefix of the file to which merged graph is written.")
    parser_merge.set_defaults(func=merge.merge_cmd)

    parser_chain = subparsers.add_parser('chain', prog="reveal chain", description="Use default chaining scheme to construct GFA graph based on a global multi-alignment of all input genomes.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])    
    parser_chain.add_argument('fastas', nargs='*', help='Fasta files specifying sequences to be (multi-)aligned into a graph.')
    parser_chain.add_argument("-o", "--output", dest="output", help="Prefix of the variant and alignment graph files to produce, default is \"sequence1_sequence2\"")
    parser_chain.add_argument("-m", dest="minlength", type=int, default=20, help="Min length of an exact (multi-)match to consider for chaining.")
    parser_chain.add_argument("-n", dest="minn", type=int, default=2, help="Only align graph on exact matches that occur in at least this many samples.")
    parser_chain.add_argument("-a", dest="maxmums", type=int, default=0, help="Number of largest mums to use for chaining (when 0 use all).")
    parser_chain.add_argument("--wp", dest="wpen", type=int, default=1, help="Multiply penalty for a MUM by this number in scoring scheme.")
    parser_chain.add_argument("--ws", dest="wscore", type=int, default=1, help="Multiply length of MUM by this number in scoring scheme.")
    parser_chain.add_argument("--gcmodel", dest="gcmodel", choices=["sumofpairs","star-avg","star-med"], default="sumofpairs", help="Which gap-cost model to use for multi-alignment.")
    parser_chain.add_argument("--recurse", dest="recurse", action="store_true", default=False, help="Use recursive approach to chain gaps.")
    parser_chain.add_argument("--plot", dest="mumplot", action="store_true", default=False, help="Save a mumplot for the actual aligned chain of anchors (depends on matplotlib).")
    parser_chain.add_argument("-i", dest="interactive", action="store_true", default=False, help="Show an interactive visualisation of the mumplot (depends on matplotlib).")
    parser_chain.add_argument("--nometa", dest="nometa", action="store_true", default=False, help="Produce a gfa graph without node annotations, to ensure it's parseable by other programs.")
    parser_chain.set_defaults(func=chain.chain_cmd)
    
    parser_stats = subparsers.add_parser('stats', prog="reveal stats", description="Output statistics (number of node, edges, genomes etc.) for a graph.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_stats.add_argument('gfa', nargs=1, help='GFA file for which statistics should be calculated.')
    parser_stats.set_defaults(func=stats.stats_cmd)

    parser_split = subparsers.add_parser('split', prog="reveal split", description="Split a graph file into a connected component per file.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, parents=[global_parser])
    parser_split.add_argument('gfa', nargs=1, help='GFA file which has to  be split into a GFA file per connected component.')
    parser_split.add_argument("--nocycles",  action="store_true", dest="nocycles", default=False, help="Parse only the directed acyclic layout of the graph, so ignore strucural rearrangements (cycles) in the graph.")
    parser_split.set_defaults(func=split.split_cmd)
    
    args = parser.parse_args()
    
    logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p', level=args.loglevel)

    args.func(args)




from utils import *
import sys


def bubbles__(G):
    stack={}
    bubblenodestack={}
    for i,node in enumerate(nx.topological_sort(G)):
        
        for (source,source_idx) in stack.keys():
            bubblenodestack[(source,source_idx)].append(node)

        nei=set(G[node].keys())
        
        if len(nei)>1: #potential source
            stack[(node,i)]=nei
            bubblenodestack[(node,i)]=[node]

        ine=[v for v,t in G.in_edges(node)]
        if len(ine)>1: #potential sink
            for (source,source_idx) in stack.keys():
                
                stack[(source,source_idx)].discard(node)
                for v in ine:
                    stack[(source,source_idx)].discard(v)
                
                # for (source,source_idx) in stack.keys():
                if stack[(source,source_idx)]==set():
                    yield Bubble(G,source,node,source_idx,i,bubblenodestack[(source,source_idx)])
                    del stack[(source,source_idx)]
                    del bubblenodestack[(source,source_idx)]
        else:
            for (source,source_idx) in stack.keys():
                for v in ine:
                    stack[(source,source_idx)].discard(v)

def bubbles_(G):
    outstack={}
    instack={}
    bubblenodestack={}
    for i,node in enumerate(nx.topological_sort(G)):
        
        for (source,source_idx) in outstack.keys():
            bubblenodestack[(source,source_idx)].append(node)

        incoming=set([v for v,t in G.in_edges(node)])
        outgoing=set([t for v,t in G.out_edges(node)])

        if len(outgoing)>1: #potential source, open a bubble
            outstack[(node,i)]=outgoing
            bubblenodestack[(node,i)]=[node]

        for source,source_idx in outstack.keys():

            outstack[(source,source_idx)].discard(node)

            if outstack[(source,source_idx)]==set():
                # print "bubble: %s <-> %s"%(source,node)
                yield Bubble(G,source,node,source_idx,i,bubblenodestack[(source,source_idx)])

                del outstack[(source,source_idx)]
                del bubblenodestack[(source,source_idx)]
            else:
                for vo in outgoing:
                    outstack[(source,source_idx)].add(vo)

def bubbles(G):
    def entrance(G,v):
        for c in G.successors(v):
            if len(list(G.predecessors(c)))==1:
                return True
        return False 

    def exit(G,v):
        for p in G.predecessors(v):
            if len(list(G.successors(p)))==1:
                return True
        return False
    
    def nextentrance(candidates,v):
        #TODO: rewrite this
        for candidate in candidates[candidates.index((v,0))+1:]:
            if candidate[1]==0:
                return candidate
    
    def superbubble(G):
        candidates=[]
        entrance2candidateidx=dict()
        sspairs=[]
        #prevEnt=None
        prevEnti=None
        alternativeEntrance={}
        previousEntrance={}

        ordD={}

        # assert(type(G)==nx.DiGraph)

        logging.debug("Topologically sort the graph.")
        ordD_=list(nx.topological_sort(G))
        logging.debug("Done.")

        #construct candidates array
        for i,v in enumerate(ordD_):
            ordD[v]=i
            alternativeEntrance[v]=None
            previousEntrance[v]=prevEnti
            if exit(G,v):
                candidates.append((v,1))
            if entrance(G,v):
                candidates.append((v,0))
                entrance2candidateidx[(v,0)]=len(candidates)-1
                prevEnti=i
        
        #construct outparent
        outparent=[None]*(len(ordD))
        for i,c in enumerate(ordD):
            tmp=[]
            for p in G.predecessors(c):
                tmp.append(ordD[p])
            if len(tmp)>0:
                outparent[ordD[c]]=min(tmp)
        
        #construct outchild
        outchild=[None]*(len(ordD))
        for i,c in enumerate(ordD):
            tmp=[]
            for p in G.successors(c):
                tmp.append(ordD[p])
            if len(tmp)>0:
                outchild[ordD[c]]=max(tmp)
        
        #loop
        while len(candidates)!=0:
            if candidates[-1][1]==0:
                del candidates[-1]
            else:
                reportsuperbubble(candidates[0],candidates[-1],candidates,previousEntrance,alternativeEntrance,G,ordD,ordD_,outchild,outparent,sspairs,entrance2candidateidx)

        return ordD,ordD_,sspairs
    
    def reportsuperbubble(vstart,vexit,candidates,previousEntrance,alternativeEntrance,G,ordD,ordD_,outchild,outparent,sspairs,entrance2candidateidx):
        
        if (vstart[0] == None) or (vexit[0] == None) or (ordD[vstart[0]] >= ordD[vexit[0]]):
            del candidates[-1]
            return

        si=previousEntrance[vexit[0]]
        
        if si==None:
            del candidates[-1]
            return

        s=ordD_[si]
        while ordD[s] >= ordD[vstart[0]]:
            valid = validatesuperbubble(s, vexit[0], ordD, ordD_, outchild, outparent, previousEntrance, G)
            
            if (valid==s):
                break

            if (valid==alternativeEntrance[s]):
                break

            if valid==-1:
                break

            alternativeEntrance[s] = valid
            s = valid

        del candidates[-1]

        if (valid == s):
            sspairs.append((s, vexit[0]))

            while (candidates[-1][0] is not s):
                if candidates[-1][1]==1:
                    
                    # ne=None
                    # for candidate in candidates[entrance2candidateidx[(s,0)]+1:]:
                    #     if candidate[1]==0:
                    #         ne=candidate
                    ne=nextentrance(candidates,s)

                    if ne!=None:
                        reportsuperbubble(ne, candidates[-1], candidates, previousEntrance, alternativeEntrance, G, ordD, ordD_, outchild, outparent,sspairs,entrance2candidateidx)
                    else:
                        del candidates[-1]
                else:
                    del candidates[-1]
    
    def validatesuperbubble(startVertex, endVertex, ordD, ordD_, outchild, outparent, previousEntrance, G):
        start=ordD[startVertex]
        end=ordD[endVertex]
        if start+1!=end:
            oc=max(outchild[start:end])
            op=min(outparent[start+1:end+1])
        else:
            oc=outchild[start]
            op=outparent[end]
        if oc!=end:
            return -1
        if op==start:
            return startVertex
        if op==None:
            return -1
        elif entrance(G, ordD_[op]):
            return ordD_[op]
        elif previousEntrance[ordD_[op]]==None: #
            return -1
        else:
            return ordD_[previousEntrance[ordD_[op]]]
        return startVertex
    
    ordD,ordD_,sspairs=superbubble(G)

    allpairs=[]
    for pair in sspairs:
        allpairs.append((pair[0],pair[1]))

    allpairs.sort(key=lambda a: ordD[a[0]])#,reverse=True) #sort by topological order of the source

    for v,u in allpairs:
        bubblenodes=ordD_[ordD[v]:ordD[u]+1]
        sourcenode=G.node[v]
        sourcesamples=set(sourcenode['offsets'].keys())
        sinknode=G.node[u]
        sinksamples=set(sinknode['offsets'].keys())

        if sinksamples!=sourcesamples:
            logging.debug("Invalid bubble, between %s and %s"%(v,u))
            continue

        if len(bubblenodes)==2: #only source sink, no variation
            continue

        yield Bubble(G,v,u,ordD[v],ordD[u],bubblenodes)

def bubbles_cmd(args):
    if len(args.graph)<1:
        logging.fatal("Specify a gfa file to extract bubbles.")
        return
    
    G=nx.DiGraph()
    read_gfa(args.graph[0],None,"",G,remap=False)

    # bubbles(G)
    # sys.exit(0)

    sys.stdout.write("#source\tsink\tsubgraph\ttype\n")
    for i,g in enumerate(nx.weakly_connected_component_subgraphs(G)):
        logging.info("Reporting bubbles for subgraph: %d"%i)
        allcomplexnodes=[]
        for b in bubbles(g):
            if type(b)!=tuple:
                t=b.issimple()
                sys.stdout.write("%s\t%s\t%s\t%s\n"%(b.source if type(b.source)!=str else '<start>',
                                                    b.sink if type(b.sink)!=str else '<end>',
                                                    ",".join([str(x) for x in b.nodes if type(x)!=str]),
                                                    'simple' if t else 'complex'))

                if not t:
                    if args.exportcomplex:
                        if args.separate:
                            sg=g.subgraph(set(b.nodes))
                            if args.gml:
                                write_gml(sg,None,outputfile=args.graph[0].replace(".gfa",".%d.%d.complex.gml"%(b.source,b.sink)),partition=False)
                            else:
                                write_gfa(sg,None,remap=False,outputfile=args.graph[0].replace(".gfa","%d.%d.complex.gfa"%(b.source,b.sink)))
                        else:
                            allcomplexnodes+=b.nodes

        if args.exportcomplex and not args.separate:
            sg=g.subgraph(allcomplexnodes)
            if args.gml:
                write_gml(sg,None,outputfile=args.graph[0].replace(".gfa",".complex.gml"),partition=False)
            else:
                write_gfa(sg,None,remap=False,outputfile=args.graph[0].replace(".gfa",".complex.gfa"))

def rearrangements_cmd(args):

    G=nx.MultiDiGraph() #if we parse a DiGraph, the edges introduced by structural variants will be ignored
    
    logging.debug("Reading graph...")
    read_gfa(args.graph[0],None,"",G)
    logging.debug("Done.")

    logging.info("Determine rearrangement edges...")
    if type(G)==nx.MultiDiGraph: #convert to DiGraph first so we can actually toposort it
        rearrangements=MultiGraphToDiGraph(G)
    logging.info("Done (%d)."%len(rearrangements))

    gori=sorted([p for p in G.graph['paths'] if not p.startswith('*')])

    if args.reference==None:
        args.reference=gori[0]

    cds=G.graph['path2id'][args.reference] if args.reference in G.graph['path2id'] else G.graph['path2id'][gori[0]]

    sys.stdout.write("#reference\tapproximate_pos\tcontigs\tsource\tsink\tinvert\tpaths\n")

    for b in rearrangements:
        v,u,k,d=b

        if type(v)==str or type(u)==str:
            continue #just start/end
        else:
            
            paths=[G.graph['id2path'][sid] for sid in d['paths']] #all paths that go through the rearrangement edge

            for p in G.node[u]['offsets'].keys():
                if G.graph['id2path'][p].startswith(args.reference):
                    vcds=p
                    break
            else:
                logging.warn("Edge %s could not be located on reference: %s."% (str((v,u)), args.reference))
                vcds=G.node[u]['offsets'].keys()[0]

            vpos=G.node[u]['offsets'][vcds]



            contigs=[]
            for p in d['paths']:
                path=G.graph['id2path'][p]
                if path.startswith("*"):
                    contigs.append(path)

            sys.stdout.write("%s\t"*7 % (G.graph['id2path'][vcds], vpos, contigs, v, u, d['oto']==d['ofrom'], ",".join(paths)))
            
            sys.stdout.write("\n")
            sys.stdout.flush()

    logging.info("Done")

def variants_cmd(args):
    if len(args.graph)<1:
        logging.fatal("Specify a gfa file to extract bubbles.")
        return
    
    reference=args.reference
    g=nx.DiGraph() #if we parse a DiGraph, the edges introduced by structural variants will be ignored
    
    logging.debug("Reading graph...")
    read_gfa(args.graph[0],None,"",g)
    logging.debug("Done.")

    complexbubblenodes=[]
    
    if 'paths' in g.graph:
        gori=sorted([p for p in g.graph['paths'] if not p.startswith('*')])
    else:
        gori=[]

    if args.reference==None:
        args.reference=gori[0]
        logging.warn("No reference specified as a coordinate system, use %s where possible."%args.reference)
        args.reference=g.graph['path2id'][args.reference]
    else:
        if args.reference in g.graph['path2id']:
            args.reference=g.graph['path2id'][args.reference]
        else:
            logging.fatal("Specified reference (%s) not available in graph, graph knows of: %s."%(args.reference,str(g.graph['paths'])))
            sys.exit(1)
    
    try:
        if not args.fastaout and not args.bedout and not args.vcfout:
            sys.stdout.write("#reference\tpos_start\tpos_end\tsource_size\tsink_size\tmax_allele_size\tmin_allele_size\tdiff_allele_size\tsource\tsink\tsource_seq\tsink_seq\ttype\tgenotypes")
            for sample in gori:
                sys.stdout.write("\t%s"%sample)
            sys.stdout.write("\n")
        elif args.vcfout:
            sys.stdout.write("##fileformat=VCFv4.0\n")#?
            sys.stdout.write("##source=REVEAL\n")
            for sid in g.graph['id2path']:
                size=g.graph['id2end'][sid]
                sys.stdout.write("##contig=<ID=%s,LENGTH=%d>\n"%(g.graph['id2path'][sid],size))
            sys.stdout.write("##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n")
            sys.stdout.write("##FORMAT=<ID=SZ,Number=1,Type=Integer,Description=\"Allele size\">\n")
            sys.stdout.write("##INFO=<ID=reveal_diffsize,Number=1,Type=Integer,Description=\"Difference between the shortest and longest allele.\">\n")
            sys.stdout.write("##INFO=<ID=reveal_source,Number=1,Type=String,Description=\"Source of the node pair.\">\n")
            sys.stdout.write("##INFO=<ID=reveal_sink,Number=1,Type=String,Description=\"Sink of the node pair.\">\n")
            sys.stdout.write("##INFO=<ID=reveal_bubbletype,Number=1,Type=String,Description=\"Simplistic interpretation of the variant.\">\n")
            sys.stdout.write("##INFO=<ID=reveal_start,Number=1,Type=String,Description=\"Start position on the specified reference.\">\n")
            sys.stdout.write("##INFO=<ID=reveal_end,Number=1,Type=String,Description=\"End position on the specified reference.\">\n")
            sys.stdout.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT")
            for sample in gori:
                sys.stdout.write("\t%s"%sample)
            sys.stdout.write("\n")

        for bi,b in enumerate(bubbles(g)):

            v=Variant(b)
            
            if v.maxsize<args.minsize:
                continue

            if v.maxsize-v.minsize<args.mindiff:
                continue

            if args.maxdiff!=None and v.maxsize-v.minsize>args.maxdiff:
                continue

            if v.vtype!=args.type and args.type!='all':
                continue
            
            genotypestr=",".join(v.genotypes)
            
            if args.nogaps:
                if v.spans_gap:
                    continue

            minflank=min([len(g.node[v.source]['seq']),len(g.node[v.sink]['seq'])])
            
            if minflank<args.minflank:
                continue

            if args.reference in v.vpos:
                cds=args.reference
            else: #source does not occur on specified reference, pick any other path that does have a location for this variant
                if args.refonly: #skip the variant if its not positionable on the reference
                    continue
                for cds in v.vpos.keys():
                    if not g.graph['id2path'][cds].startswith('*'): #use ref layout if its there
                        break

            sourcelen=len(g.node[v.source]['seq'])
            sinklen=len(g.node[v.sink]['seq'])
            
            startpos=g.node[v.source]['offsets'][cds]+sourcelen
            endpos=g.node[v.sink]['offsets'][cds]

            if args.fastaout:
                if args.split:
                    with open("%s_%s.fasta"%(v.source,v.sink),'w') as of:
                        for i,seq in enumerate(v.genotypes):
                            if seq!='-':
                                of.write(">%s:%d-%d_%d\n"%(g.graph['id2path'][cds],startpos,endpos,i))
                                of.write("%s\n"%seq)
                else:
                    for i,seq in enumerate(v.genotypes):
                        if seq!='-':
                            sys.stdout.write(">%s:%d-%d_%d\n"%(g.graph['id2path'][cds],startpos,endpos,i))
                            sys.stdout.write("%s\n"%seq)
                continue

            if args.bedout:
                sys.stdout.write("%s\t%d\t%s\t%s\n"%(g.graph['id2path'][cds],startpos,endpos,v.vtype))
                continue

            allelesizes=[]

            for gt in v.genotypes:
                if gt=='-':
                    allelesizes.append(0)
                else:
                    allelesizes.append(len(gt))
            
            maxa=max(allelesizes)
            mina=min(allelesizes)

            if args.vcfout:
                startpos+=1
                if maxa-mina>0:
                    startpos-=1
                    genotypes=[]
                    for gt in v.genotypes:
                        if gt=='-':
                            gt=""
                        genotypes.append(g.node[v.source]['seq'][-1:]+gt)
                    v.genotypes=genotypes

                if v.calls[g.graph['id2path'][cds]]!=0: #for vcf output flip alleles to make reference allele 0
                    v.genotypes[0],v.genotypes[v.calls[g.graph['id2path'][cds]]]=v.genotypes[v.calls[g.graph['id2path'][cds]]],v.genotypes[0]
                _calls=dict()
                for sample in v.calls:
                    if v.calls[sample]==v.calls[g.graph['id2path'][cds]]: #same allele as ref, so make 0
                        _calls[sample]=0
                    elif v.calls[sample]==0:
                        _calls[sample]=v.calls[g.graph['id2path'][cds]]
                    else:
                        _calls[sample]=v.calls[sample]
                v.calls=_calls

                sys.stdout.write("%s\t%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s"% (g.graph['id2path'][cds],
                                                                    startpos,
                                                                    ".",
                                                                    v.genotypes[0],
                                                                    ",".join(v.genotypes[1:]),
                                                                    ".",
                                                                    "PASS",
                                                                    "reveal_diffsize=%s;reveal_source=%s;reveal_sink=%s;reveal_bubbletype=%s;reveal_start=%d;reveal_end=%d"%(maxa-mina, 
                                                                                                            v.source if type(v.source)!=str else '<start>', 
                                                                                                            v.sink if type(v.sink)!=str else '<end>',
                                                                                                            v.vtype,
                                                                                                            startpos,
                                                                                                            endpos),
                                                                    "GT:SZ"
                                                                    ))

                for sample in gori:
                    if sample in v.calls:
                        sys.stdout.write("\t%s:%d"%(v.calls[sample], len(v.genotypes[v.calls[sample]])))
                    else:
                        sys.stdout.write("\t.")

            else:
                sys.stdout.write("%s\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%s\t%s\t%s\t%s\t%s\t%s"% (g.graph['id2path'][cds],
                                                                    startpos,
                                                                    endpos,
                                                                    sourcelen,
                                                                    sinklen,
                                                                    maxa,
                                                                    mina,
                                                                    maxa-mina,
                                                                    v.source if type(v.source)!=str else '<start>',
                                                                    v.sink if type(v.sink)!=str else '<end>',
                                                                    g.node[v.source]['seq'][-20:] if v.source in g else '-',
                                                                    g.node[v.sink]['seq'][:20] if v.sink in g else '-',
                                                                    v.vtype,
                                                                    genotypestr))
                for sample in gori:
                    if sample in v.calls:
                        sys.stdout.write("\t%s"%v.calls[sample])
                    else:
                        sys.stdout.write("\t-")
            
            sys.stdout.write("\n")

            sys.stdout.flush()
    except IOError:
        pass

class InvalidBubble(Exception):
    pass

class Bubble:
    def __init__(self,G,source,sink,source_idx,sink_idx,nodes):
        self.source=source
        self.sink=sink
        self.source_idx=source_idx
        self.sink_idx=sink_idx
        self.G=G
        self.nodes=nodes
        self.ordD={node:i for i,node in enumerate(nodes)}

        if len(self.nodes)<=2:
            raise InvalidBubble("Not a valid source sink pair as bubble")

        self.simple=None
        
        self.paths=set([k for k in G.node[self.source]['offsets'].keys() if not G.graph['id2path'][k].startswith("*")]) & set([k for k in G.node[self.sink]['offsets'].keys() if not G.graph['id2path'][k].startswith("*")])

        if 'seq' in G.node[self.source]:
            l=len(G.node[self.source]['seq'])
        else:
            l=(self.source[1]-self.source[0])

        self.allelesizes=[G.node[self.sink]['offsets'][p]-(G.node[self.source]['offsets'][p]+l) for p in self.paths]

        self.minsize=min(self.allelesizes)
        # self.minsize=min([len(G.node[node]['seq']) for node in self.nodes[1:-1]])

        assert(self.minsize>=0)

        # self.maxsize=max([len(G.node[node]['seq']) for node in self.nodes[1:-1]])
        self.maxsize=max(self.allelesizes)

        self.cumsize=sum(self.allelesizes)
        # self.cumsize=sum([len(G.node[node]['seq'])*len(G.node[node]['offsets']) for node in self.nodes[1:-1]])
    
    def issimple(self):
        if self.simple==None:
            
            sucs=set(self.G.successors(self.source))
            pres=set(self.G.predecessors(self.sink))
            
            sucs.discard(self.sink)
            pres.discard(self.source)
            
            for suc in sucs:
                if len(list(self.G.successors(suc)))!=1 or list(self.G.successors(suc))[0]!=self.sink:
                    self.simple=False
                    return self.simple
            
            for pre in pres:
                if len(list(self.G.predecessors(pre)))!=1 or list(self.G.predecessors(pre))[0]!=self.source:
                    self.simple=False
                    return self.simple
            
            self.simple=True
            
            return self.simple
        else:
            return self.simple

    #returns the amount the left and right margin for indel positioning
    def getwiggle(self,minwiggle=0):
        if self.issimple():

            if self.G.has_edge(self.source,self.sink):
                #how far can we move this bubble to the right?

                if 'seq' in self.G.node[self.sink]:
                    sink=self.G.node[self.sink]['seq']
                else:
                    sink=""

                if 'seq' in self.G.node[self.source]:
                    source=self.G.node[self.source]['seq']
                else:
                    source=""

                vs=[self.G.node[n]['seq']+sink for n in self.nodes[1:-1]]
                lvs=[len(s) for s in vs]+[len(sink)]
                i=0
                while i<min(lvs) and sink[i]==vs[0][i]:
                    for v in vs[1:]:
                        if not v[i]==sink[i]:
                            break
                    i+=1
                
                vs=[source+self.G.node[n]['seq'] for n in self.nodes[1:-1]]
                lvs=[len(s) for s in vs]+[len(source)]
                j=1
                while j<=min(lvs) and source[-j]==vs[0][-j]:
                    for v in vs[1:]:
                        if not v[-j]==source[-j]:
                            break
                    j+=1

                return (minwiggle+j-1,minwiggle+i) #tuple with margin on the left and margin on the right
        
        return (minwiggle,minwiggle)

class Variant(Bubble):
    def __init__(self,bubble):

        Bubble.__init__(self,bubble.G,bubble.source,bubble.sink,bubble.source_idx,bubble.sink_idx,bubble.nodes)
        
        self.genotypes=[] #list of variant sequence
        self.vtype='undefined' #type definition of the variant
        self.gap='undefined'
        self.calls=dict() #key is sample, value is index within genotypes
        self.vpos=dict() #key is sample, value is position within sample
        self.spans_gap=False

        gt=list(set(self.G.successors(self.source)) & set(self.nodes))
        gt.sort(key=lambda l: self.ordD[l])
        bsamples=set(self.G.node[self.source]['offsets'].keys())&set(self.G.node[self.sink]['offsets'].keys())

        # bsamplestmp=bsamples.copy()
        # if self.issimple():
        #     for i,v in enumerate(gt):
        #         if v==self.sink:
        #             self.genotypes.append('-')
        #         else:
        #             s=self.G.node[v]['seq']
        #             self.genotypes.append(s)

        #         for sampleid in self.G.node[v]['offsets'].keys():
        #             if sampleid in bsamplestmp:
        #                 self.calls[bubble.G.graph['id2path'][sampleid]]=i
        #                 bsamplestmp.discard(sampleid)
        # else:
        
        self.vtype="complex"

        seqd=dict()
        for sid in bsamples:
            seq=""
            for v in self.nodes[1:-1]: #determine sequence through the complex bubble; use the entire path as genotype
                if sid in self.G.node[v]['offsets']:
                    seq+=self.G.node[v]['seq']

            if seq=="":
                seq="-"

            if seq in seqd:
                seqd[seq].append(sid)
            else:
                seqd[seq]=[sid]

        self.genotypes=list(seqd.keys())
        for i,k in enumerate(self.genotypes):
            for sid in seqd[k]:
                self.calls[bubble.G.graph['id2path'][sid]]=i
        
        if self.issimple():
            if self.G.has_edge(self.source,self.sink):
                self.vtype='indel'
            elif len(self.genotypes)==2:
                if len(self.genotypes[0])==1 and len(self.genotypes[1])==1:
                    self.vtype='snp'
                else:
                    self.vtype='region'
            else:
                self.vtype='multi-allelic'

        for node in self.nodes:
            if 'N' in self.G.node[node]['seq']:
                self.spans_gap=True
                if self.issimple():
                    self.gap="simplegap"
                else:
                    self.gap="complexgap"
                break

        v=self.G.node[self.source]
        t=self.G.node[self.sink]
        o=set(v['offsets'].keys())&set(t['offsets'].keys())
        for s in o:
            self.vpos[s]=v['offsets'][s]+len(v['seq'])+1


import networkx as nx
from utils import *
from intervaltree import IntervalTree
import bubbles

def subgraph(args):
    if len(args.inputfiles)<=1:
        logging.fatal("Specify 1 gfa file followed by comma separated list of node ids that make up the subgraph.")
        return

    if not args.inputfiles[0].endswith('.gfa'):
        logging.fatal("Specify gfa file as first argument of subgraph subcommand.")
        return

    G=nx.DiGraph()

    read_gfa(args.inputfiles[0],None,"",G)
    
    topsort=list(nx.topological_sort(G))

    nodes=set()
    
    for arg in args.inputfiles[1:]:
        if arg.find(':')!=-1: #then parse as interval definition: eg chr4:230000-230010
            cds,intv=arg.split(':')
            tree=graph_to_tree(G,cds) #use interval tree for retrieval
            start,stop=intv.split('-')
            
            source=tree[int(start)].pop()[2]
            sink=tree[int(stop)].pop()[2]
            source_idx,sink_idx=0,0
            for i,v in enumerate(topsort):
                if v==source:
                    source_idx=i
                if v==sink:
                    sink_idx=i
                    break
            if sink_idx<=source_idx:
                logging.fatal("Invalid source/sink pair. Topsort indices: %d, %d"%(source_idx,sink_idx))
                return

            for node in topsort[source_idx:sink_idx+1]:
                nodes.add(int(node))

        elif arg.find('-')!=-1: #then bubble definition
            source,sink=arg.split('-')

            for i,v in enumerate(topsort):
                if v==int(source):
                    source_idx=i
                if v==int(sink):
                    sink_idx=i
                    break
            else:
                logging.fatal("Sink node could not be found.")
                sys.exit(1)

            if sink_idx<=source_idx:
                logging.fatal("Invalid source/sink pair. Topsort indices: %d, %d"%(source_idx,sink_idx))
                return
            
            for node in topsort[source_idx:sink_idx+1]:
                nodes.add(int(node))
        else:
            for node in arg.split(','): #assume a comma separated list of nodes
                nodes.add(int(node))
    
    sg=G.subgraph(nodes)
    sg.graph['startnodes']=[]

    for sid in sg.graph['id2path']:
        start=None
        for node in sg:
            if sid in sg.node[node]['offsets']:
                if start==None or sg.node[node]['offsets'][sid]<start:
                    start=sg.node[node]['offsets'][sid]
                    startnode=node
        sg.graph['startnodes'].append(startnode)

    # for sid in sg.graph['id2path']:
    #     end=None
    #     for node in sg:
    #         if sid in sg.node[node]['offsets']:
    #             if end==None or sg.node[node]['offsets'][sid]>end:
    #                 end=sg.node[node]['offsets'][sid]
    #                 endnode=node
    #     sg.graph['endnodes'].append(endnode)

    if args.gml:
        write_gml(sg,"",outputfile=args.outfile)
    else:
        write_gfa(sg,"",outputfile=args.outfile, remap=False)

def graph_to_tree(G,cds):
    tree=IntervalTree()
    cds=G.graph['path2id'][cds]
    for node in G:
        if type(node)!=str:
            if cds in G.node[node]['offsets']:
                tree[G.node[node]['offsets'][cds]:G.node[node]['offsets'][cds]+len(G.node[node]['seq'])]=node
    return tree



import sys
import os
import logging
import uuid
import gzip

def annotate(args):
    import pysam

    vcf_reader=pysam.VariantFile(args.vcffile,"r")

    aid=0
    aid2variant={}
    vi=0

    vfile=args.vcffile+".fasta"
    with open(vfile,'w') as v:
        for record in vcf_reader:
            if 'reveal_diffsize' in record.info:
                if record.info['reveal_diffsize']>=args.mindiff and record.info['reveal_diffsize']<args.maxdiff:
                    for i,allele in enumerate(record.alleles):
                        allele=str(allele)
                        aid+=1
                        aid2variant[aid]=(record.chrom,record.pos,i,len(allele))
                        # variant2aid[(record.CHROM,record.POS,i)]=aid
                        # v.write(">%s\n"%str(record.CHROM+"_"+str(record.POS)+"_"+str(i)))
                        v.write(">%d\n"%aid)
                        for i in range((len(allele)/50)+1): #write in blocks of 50
                            v.write("%s\n"% allele[i*50:(i+1)*50] )
                    vi+=1

    if vi==0:
        logging.info("No variants in this size range.")
        sys.exit(1)
    
    vcf_reader=pysam.VariantFile(args.vcffile,"r")

    repmd={}
    if args.repm:
        #call repeatmasker
        logging.info("Running repeatmasker...")
        if os.system("RepeatMasker -species %s -pa %d -nolow -nocut %s"%(args.species,args.repmproc,vfile))!=0:
            logging.fatal("RepeatMasker failed, make sure calls to \'RepeatMasker\' are possible.")
            sys.exit(1)
        logging.info("Done.")

        repmd=load_repm_annotations(vfile+".out",aid2variant)

        #clean up
        for ext in [".cat",".cat.all",".out",".tbl",".masked"]:
            try:
                os.remove(vfile+ext)
            except OSError:
                pass

        if 'repm_rfamily' not in vcf_reader.header.info:
            vcf_reader.header.info.add('repm_rfamily', 1, 'String', 'Best match for RepeatMasker - Repeat family.')
        if 'repm_rtype' not in vcf_reader.header.info:
            vcf_reader.header.info.add('repm_rtype', 1, 'String', 'Best match for RepeatMasker - Repeat type.')
        if 'repm_rcov' not in vcf_reader.header.info:
            vcf_reader.header.info.add('repm_rcov', 1, 'Float', 'Best match for RepeatMasker - Fraction of the repeat annotation covered.')
        if 'repm_qcov' not in vcf_reader.header.info:
            vcf_reader.header.info.add('repm_qcov', 1, 'Float', 'Best match for RepeatMasker - Fraction of the indel covered.')
        if 'repm_allele' not in vcf_reader.header.info:
            vcf_reader.header.info.add('repm_allele', 1, 'Integer', 'The allele that contains the best RepeatMasker match.')

    trfd={}
    if args.trf:
        #call trf
        logging.info("Running tandem repeat finder...")
        if os.system("trf %s 2 7 7 80 10 20 500 -ngs -h > %s.trf"%(vfile,vfile))!=0:
            logging.fatal("Tandem Repeat Finder failed, make sure calls to \'trf\' are possible.")
            sys.exit(1)
        logging.info("Done.")

        trfd=load_trf_annotations(vfile+".trf",aid2variant)

        #clean up
        for ext in [".trf"]:
            try:
                os.remove(vfile+ext)
            except OSError:
                pass

        if 'trf_copynumber' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_copynumber', 1, 'Float', 'Best match for TRF - copynumber.')
        if 'trf_conssize' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_conssize', 1, 'Integer', 'Best match for TRF - concensus size.')
        if 'trf_entropy' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_entropy', 1, 'Float', 'Best match for TRF - entropy.')
        if 'trf_pattern' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_pattern', 1, 'String', 'Best match for TRF - pattern.')
        if 'trf_start' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_start', 1, 'String', 'Best match for TRF - start position of tr.')
        if 'trf_end' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_end', 1, 'String', 'Best match for TRF - end position of tr.')
        if 'trf_gccontent' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_gccontent', 1, 'Float', 'Best match for TRF - Fraction of GC bases in the repeat pattern.')
        if 'trf_percent_indel' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_percent_indel', 1, 'Integer', 'Best match for TRF - percentage of indels within the aligned repeat pattern.')
        if 'trf_percent_match' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_percent_match', 1, 'Integer', 'Best match for TRF - percentage of matches within the aligned repeat pattern.')
        if 'trf_cov' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_cov', 1, 'Float', 'Best match for TRF - fraction of the allele that is masked by this tandem repeat pattern')
        if 'trf_allele' not in vcf_reader.header.info:
            vcf_reader.header.info.add('trf_allele', 1, 'Integer', 'The allele that contains the best TRF match.')

    #clean up
    os.remove(vfile)
    
    try: #if its already there, leave, just try to relabel
        vcf_reader.header.info.add('reveal_type', 1, 'String', 'REVEAL\'s best guess at the type of variant.')
    except:
        pass

    if args.vcffile.endswith('.gz'):
        outputfile=gzip.open(args.vcffile[:-7]+".annotated"+args.vcffile[-7:],'w')
    else:
        outputfile=open(args.vcffile[:-4]+".annotated"+args.vcffile[-4:],'w')

    vcf_writer = pysam.VariantFile(outputfile, 'w', header=vcf_reader.header)

    try:
        for record in vcf_reader:

            if 'reveal_diffsize' in record.info:
                if record.info['reveal_diffsize']>=args.mindiff and record.info['reveal_diffsize']<args.maxdiff:
                    key=(record.chrom,record.pos)

                    if key in repmd: #we have a repeat masker annotation for this allele
                        record.info['repm_rfamily']=repmd[key]['rfamily']
                        record.info['repm_rtype']=repmd[key]['rtype']
                        record.info['repm_rcov']=repmd[key]['rcov']
                        record.info['repm_qcov']=repmd[key]['qcov']
                        record.info['repm_allele']=repmd[key]['allele'] #numeric representation of the allele that this annotation was based on

                    if key in trfd: #we have a trf annotation for this allele
                        record.info['trf_copynumber']=trfd[key]['copynumber']
                        record.info['trf_conssize']=trfd[key]['cons_size']
                        record.info['trf_entropy']=trfd[key]['entropy']
                        record.info['trf_start']=trfd[key]['start']
                        record.info['trf_end']=trfd[key]['end']
                        record.info['trf_pattern']=trfd[key]['pattern']
                        record.info['trf_percent_indel']=trfd[key]['percent_indel']
                        record.info['trf_percent_match']=trfd[key]['percent_match']
                        record.info['trf_gccontent']=(trfd[key]['G']+trfd[key]['C'])/float(trfd[key]['A']+trfd[key]['C']+trfd[key]['G']+trfd[key]['T'])
                        record.info['trf_cov']=len(trfd[key]['masked'])/float(trfd[key]['allelesize'])
                        record.info['trf_allele']=trfd[key]['allele'] #numeric representation of the allele that this annotation was based on

                    #add custom reveal annotation derived from repeatmasker and trf annotations
                    if key in repmd and repmd[key]['rcov']>0.8 and repmd[key]['qcov']>0.8 and not repmd[key]['rfamily'].startswith('Satellite'): #repeat annotation and allele have reciprocal overlap >0.8
                        record.info['reveal_type']='mei'
                    elif key in trfd and record.info['trf_cov']>0.5: #no mei, but more than 50% of the indel size can be attributed to tandemly repeated sequence
                        if trfd[key]['cons_size']==1:
                            record.info['reveal_type']='homopolymer'
                        elif trfd[key]['cons_size']<=6:
                            record.info['reveal_type']='micro-satellite'
                        elif trfd[key]['cons_size']<100:
                            record.info['reveal_type']='mini-satellite'
                        elif trfd[key]['cons_size']<1000:
                            record.info['reveal_type']='macro-satellite'
                        elif trfd[key]['cons_size']>1000:
                            record.info['reveal_type']='mega-satellite'
                    else:
                        record.info['reveal_type']='other'

            vcf_writer.write(record)

        vcf_writer.close()
        
    except IOError:
        vcf_writer.close()


def load_trf_annotations(trffile,aid2variant):
    trfd={}
    pvariant=None
    trfcolnames=['start','end','period_size','copynumber','cons_size','percent_match','percent_indel','score','A','C','G','T','entropy','pattern','masked']
    trfcoltypes=[str,str,int,float,int,int,int,int,int,int,int,int,float,str,str]
    #13 47 1 35.0 1 100 0 70 0 0 0 100 0.00 T TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT

    with open(trffile) as trfout:
        for line in trfout:
            line=line.strip()
            if line[0]=='@':
                aid=int(line[1:])
                allele=aid2variant[aid][2]
                allelesize=aid2variant[aid][3]
                variant=aid2variant[aid][:2]
                continue
            else:
                cols=line.rstrip().split()
                repeat={k:t(v) for k,t,v in zip(trfcolnames[:15],trfcoltypes[:15],cols[:15])}
                repeat['allelesize']=allelesize
                repeat['allele']=allele
            
            if pvariant==variant:
                if repeat['score']>trfd[variant]['score']:
                    trfd[variant]=repeat
            else:
                trfd[variant]=repeat
            
            pvariant=variant

    return trfd


def load_repm_annotations(repmfile,aid2variant):
    repmd={}
    pvariant=None

    repmcolnames=['score','div','del','ins','qsequence','qbegin','qend','qleft','C_','rtype','rfamily','rbegin','rend','rleft','vid','rcov','qcov']
    with open(repmfile) as repmfile:
        for i in range(3):
            h=repmfile.readline()
        
        for line in repmfile:
            cols=line.split()
            
            repeat={k:v for k,v in zip(repmcolnames,cols)}
            
            repeat['score']=float(repeat['score'])
            
            aid=int(cols[4])
            allele=aid2variant[aid][2]
            variant=aid2variant[aid][:2]
            
            qcov=(int(repeat['qend'])-int(repeat['qbegin']))/float(int(repeat['qleft'][1:-1])+int(repeat['qend']))
            
            if repeat['rbegin'][0]=='(':
                rcov=(int(repeat['rend'])-int(repeat['rleft']))/float(int(repeat['rbegin'][1:-1])+int(repeat['rend']))
            else:
                rcov=(int(repeat['rend'])-int(repeat['rbegin']))/float(int(repeat['rleft'][1:-1])+int(repeat['rend']))
                    
            assert(rcov>0 and rcov<=1)
            assert(qcov>0 and qcov<=1)
            
            repeat['rcov']=rcov
            repeat['qcov']=qcov
            repeat['allele']=allele

            if pvariant==variant:
                if repeat['score']>repmd[variant]['score']:
                    repmd[variant]=repeat
            else:
                repmd[variant]=repeat
            
            pvariant=variant

    return repmd
















import utils
import bubbles
import networkx as nx
import os
import logging
import sys

def chop_cmd(args):
    if not args.graph[0].endswith(".gfa"):
        logging.fatal("Invalid gfa file.")
        return

    G=nx.DiGraph()
    utils.read_gfa(args.graph[0], None, None, G, remap=False)
    
    # assert(len(G.edges())==0)
    if args.output==None:
        fof=os.path.splitext(args.graph[0])[0]+".chopped.fasta"
        gof=os.path.splitext(args.graph[0])[0]+".chopped.gfa"
    else:
        fof=args.output+".fasta"
        gof=args.output+".gfa"

    if args.check:
        Gorg=G.copy()

    chop(G,k=args.k,extend=args.extend)
    
    logging.debug("Merging node sequence...")
    for node in G.nodes():
        if type(node)==str: #skip start/end nodes
            continue
        G.node[node]['seq']=G.node[node]['prefix']+G.node[node]['seq']+G.node[node]['suffix']
    
    logging.debug("Done.")

    logging.debug("Write overlap graph...")
    utils.write_gfa(G,None,outputfile=gof,remap=False)
    logging.debug("Done.")

    if args.fasta:
        logging.debug("Write corresponding fasta file...")
        with open(fof,'w') as ff:
            for node in G.nodes():
                if type(node)==str: #skip start/end nodes
                    continue
                name=">"+str(node)+"\n"
                seq=G.node[node]['seq']
                ff.write(name)
                for i in range( (len(seq)/args.lw)+(len(seq) % args.lw > 0)):
                    ff.write(seq[i*args.lw:(i+1)*args.lw]+"\n")
        logging.debug("Done.")
    
    if args.check:
        logging.debug("Validate if all k-mers from the original graph are contained in overlap graph...")
        import extract
        r="$".join([G.node[node]['seq'] for node in G])
        for path in Gorg.graph['paths']:
            logging.debug("Check: %s"%path)
            s=extract.extract(Gorg,path)
            for i in xrange(len(s)-args.k):
                if r.find(s[i:i+args.k])==-1:
                    logging.error("Flat representation does not cover all k-length substrings for %s, could not find: %s!"%(path,s[i:i+args.k]))
                    sys.exit(1)
        logging.debug("Done.")

def duplicate_node(G,node):
    if type(node)==str:
        logging.fatal("Attempt to duplicate end/start node, shouldn't happen. Exit.")
        sys.exit(1)
    logging.debug("Duplicate: node %s"%node)
    offsets=G.node[node]['offsets']
    seq=G.node[node]['seq']
    es=[]
    duplicates=[]
    predecessors=list(G.predecessors(node))
    successors=list(G.successors(node))
    if len(predecessors)>0 and len(successors)>0:
        for pred in predecessors:
            for suc in successors:
                i=G[pred][node]['paths'].intersection(G[node][suc]['paths'])
                if len(i)>0:
                    G.add_node(G.graph['noffset'],offsets={k:offsets[k] for k in offsets if k in i},seq=seq,prefix="",suffix="") #TODO: prevent contract, by doing so immediately
                    duplicates.append(G.graph['noffset'])
                    es.append((pred,G.graph['noffset'],{'paths':i,'ofrom':G[pred][node]['ofrom'],'oto':G[pred][node]['oto']}))
                    es.append((G.graph['noffset'],suc,{'paths':i,'ofrom':G[node][suc]['ofrom'],'oto':G[node][suc]['oto']}))
                    G.graph['noffset']+=1
    elif len(predecessors)>0:
        for pred in predecessors:
            i=G[pred][node]['paths']
            G.add_node(G.graph['noffset'],offsets={k:offsets[k] for k in offsets if k in i},seq=seq,prefix="",suffix="") #TODO: prevent contract, by doing so immediately
            duplicates.append(G.graph['noffset'])
            es.append((pred,G.graph['noffset'],{'paths':i,'ofrom':G[pred][node]['ofrom'],'oto':G[pred][node]['oto']}))
            G.graph['noffset']+=1
    elif len(successors)>0:
        for suc in successors:
            i=G[node][suc]['paths']
            G.add_node(G.graph['noffset'],offsets={k:offsets[k] for k in offsets if k in i},seq=seq,prefix="",suffix="") #TODO: prevent contract, by doing so immediately
            duplicates.append(G.graph['noffset'])
            es.append((G.graph['noffset'],suc,{'paths':i,'ofrom':G[node][suc]['ofrom'],'oto':G[node][suc]['oto']}))
            G.graph['noffset']+=1
    
    G.remove_node(node)
    G.add_edges_from(es)

    return duplicates


def checkedges(G,k=100):
    for u,v,d in G.edges(data=True):
        d['overlap']=None
    es=[]
    
    update=True
    while update:
        update=False
        
        remove=[]
        for u,v,d in G.edges(data=True):
            if d['overlap']!=None:
                continue

            if type(u)==str:
                d['overlap']=u
                continue

            if type(v)==str:
                d['overlap']=v
                continue

            if len(G.node[u]['seq'])>=k-1 and len([e for e in G.in_edges(v)])==1:
                d['overlap']=u
                update=True
                continue #can use k-1 suffix of u as prefix of v

            if len(G.node[v]['seq'])>=k-1 and len([e for e in G.out_edges(u)])==1:
                d['overlap']=v
                update=True
                continue #can use k-1 prefix of v as suffix of u

            #substitute edge by node that contains only pre- and suffix
            if len(G.node[v]['seq'])>=k-1 and len(G.node[u]['seq'])>=k-1:
                nid=G.graph['noffset']
                G.add_node(nid,seq="",offsets=G.nodes[v]['offsets'],prefix="",suffix="")
                G.graph['noffset']+=1
                G.add_edge(u,nid,**G[u][v])
                G[u][nid]['overlap']=u
                G.add_edge(nid,v,**G[u][v])
                G[nid][v]['overlap']=v
                remove.append((u,v))
                update=True

            # if len(G.node[u]['seq'])>=k-1 and len([e for e in G.in_edges(v,data=True) if e[2]['overlap']==None])==1:
            #     d['overlap']=u
            #     update=True
            #     continue #can use k-1 suffix of u as prefix of v

            # if len(G.node[v]['seq'])>=k-1 and len([e for e in G.out_edges(u,data=True) if e[2]['overlap']==None])==1:
            #     d['overlap']=v
            #     update=True
            #     continue #can use k-1 prefix of v as suffix of u

        G.remove_edges_from(remove)

    for u,v,d in G.edges(data=True):
        # if type(u)==str or type(v)==str:
        #     continue
        if d['overlap']==None:
            es.append((u,v))

    return es

def chop(G,k=100,extend=True):
    # remove=[]
    for node in G.nodes():
        if type(node)==str:
            pass
            # remove.append(node)
        else: #add prefix and suffix attributes
            G.node[node]['prefix']=""
            G.node[node]['suffix']=""

    # G.remove_nodes_from(remove)
    iteration=1

    es=checkedges(G,k=k)

    maxiter=1e22

    while len(es)!=0 and iteration<maxiter:
        logging.info("Running iteration %d"%iteration)
        
        #determine subgraph for duplication
        sg=nx.DiGraph(es)

        nodes=list(sg.nodes())

        nodes=[node for node in nodes if (len(sg.in_edges(node))>1 or len(sg.out_edges(node))>1) and type(node)!=str]# and len(G.node[node]['seq'])<k-1]
        nodes.sort(key=lambda n: len(G.node[n]['seq']))
        d=set()
        dups=[]
        for n in nodes:
            dup=True
            for n1,n2 in sg.in_edges(n):
                if n1 in d:
                    dup=False
                d.add(n1)
            for n1,n2 in sg.out_edges(n):
                if n2 in d:
                    dup=False
                d.add(n2)
            if dup:
                dups.append(n)
        
        logging.info("Duplicating nodes...")
        for n in dups:
            logging.debug("Duplicating node %d..."%n)
            for dup in duplicate_node(G,n):
                logging.debug("Generated node %d"%dup)
        logging.info("Duplicating done.")

        logging.info("Contracting nodes...")
        
        topsort=list(nx.topological_sort(G))[1:-1]
        topsort=[v for v in topsort if type(v)!=str]
        
        utils.contract(G,topsort)
        logging.info("Contracting done.")

        # es=checkedges(G,k=k)
        # sg=nx.DiGraph(es)
        # for u,v,d in sg.edges(data=True):
        #     if len(sg.in_edges(v))==len(sg.out_edges(u))==1 and len(G.nodes[u]['seq'])>=k-1 and len(G.nodes[v]['seq'])>=k-1:
        #         nid=G.graph['noffset']
        #         assert(nid not in G)
        #         G.add_node(nid,seq="",offsets=G.nodes[v]['offsets'],prefix="",suffix="")
        #         G.graph['noffset']+=1
        #         G.add_edge(u,nid,**G[u][v])
        #         G.add_edge(nid,v,**G[u][v])
        #         G.remove_edge(u,v)

        logging.debug("Checking edges...")
        es=checkedges(G,k=k)

        logging.info("Done. %d unextendable edges remain."%len(es))

        for u,v in es:
            logging.debug("Edge %s,%s can't be extended yet: %s"%(u,v,G[u][v]))

        iteration+=1

    if len(es)>0:
        logging.fatal("Error, maxiterations reached, chop did not converge!")
        sys.exit(1)

    if extend:
        logging.info("Extending nodes with prefix/suffix...")

        #all edges can now be extended
        for u,v,d in G.edges(data=True):
            if type(u)==str or type(v)==str:
                continue

            assert(d['overlap']!=None)
            
            if d['overlap']==u:
                logging.debug("Add prefix to %s"%v)
                assert(G.node[v]['prefix']=="")
                G.node[v]['prefix']=G.node[u]['seq'][-(k-1):]
                #if we give v a prefix, all other incoming edges of v are also affected, so cigar should be increased on those as well 
                # for _u,_v,_d in G.in_edges(v,data=True):
                    # _d['overlap_length']+=(k-1)
            else:
                assert(d['overlap']==v)
                logging.debug("Add suffix to %s"%u)
                assert(G.node[u]['suffix']=="")
                G.node[u]['suffix']=G.node[v]['seq'][:k-1]
                #if we give u a suffix, all other outgoing edges of u are also affected, so cigar should be increased on those as well
                # for _u,_v,_d in G.out_edges(u,data=True):
                    # _d['overlap_length']+=(k-1)

            d['cigar']=str(k-1)+"M"

        logging.info("Done.")
    
    # logging.info("Unzipping bubbles...")
    # #all edges can now be extended
    # for n in G.nodes():
    #     extend=False
    #     for u,v,d in G.out_edges(n,data=True):
    #         if d['overlap']==v: #overlap comes from v, so use prefix of v as suffix for u
    #             if extend:
    #                 logging.error("PROBLEM out!")
    #                 print n,u,v,d
    #             extend=True
    #     prefix=False
    #     for u,v,d in G.in_edges(n,data=True):
    #         if d['overlap']==u: #overlap comes from u, so use suffix of u as prefix for v
    #             if prefix:
    #                 logging.error("PROBLEM in!")
    #                 print n,u,v,d
    #             prefix=True

    return G



from utils import *
import reveallib
import reveallib64
import os


def matches(args): 
    
    if args.sa64:
        idx=reveallib64.index(sa=args.sa1, lcp=args.lcp1, cache=args.cache) #enable preconstruction of first SA and LCP array
    else:
        idx=reveallib.index(sa=args.sa1, lcp=args.lcp1, cache=args.cache) #enable preconstruction of first SA and LCP array
    
    G=nx.DiGraph()
    G.graph['paths']=[]
    t=IntervalTree()
    
    reffile=os.path.basename(args.reference)
    ctgfile=os.path.basename(args.contigs)
    
    ref2length=dict()
    idx.addsample(reffile)
    if args.reference.endswith(".gfa"):
        read_gfa(args.reference,idx,t,G)
    else:
        G.graph['paths'].append(reffile)
        for name,seq in fasta_reader(args.reference):
            ref2length[name]=len(seq)
            intv=idx.addsequence(seq)
            intv=Interval(intv[0],intv[1],name)
            t.add(intv)
            G.add_node(intv,offsets={reffile:0})
    
    contig2length=dict()
    idx.addsample(ctgfile)
    if args.contigs.endswith(".gfa"):
        read_gfa(args.contigs,idx,t,G)
    else:
        G.graph['paths'].append(ctgfile)
        for name,seq in fasta_reader(args.contigs):
            contig2length[name]=len(seq)
            intv=idx.addsequence(seq)
            intv=Interval(intv[0],intv[1],name)
            t.add(intv)
            G.add_node(intv,offsets={ctgfile:0})
    
    #map nodes to connected components in the graph
    refnode2component=dict()
    ctgnode2component=dict()
    component2refnode=dict()
    component2ctgnode=dict()
    refcomponents=[]
    ctgcomponents=[]
    ctg2ref=dict()
    ri=0
    ci=0
    for nodes in nx.connected_components(G.to_undirected()):
        nodes=list(nodes)
        if reffile in G.node[nodes[0]]['offsets']:
            for node in nodes:
                assert(reffile in G.node[node]['offsets']) #check the graph is valid
                refnode2component[node]=ri
                component2refnode[ri]=node
            ri+=1
            refcomponents.append(nodes)
        else:
            for node in nodes:
                assert(ctgfile in G.node[node]['offsets']) #check the graph is valid
                ctgnode2component[node]=ci
                component2ctgnode[ci]=node
            ci+=1
            ctgcomponents.append(nodes)
    
    #for each contig, print the length
    for name in contig2length:
        print "#%s\t%d"%(name,contig2length[name])
    
    idx.construct()
    
    if args.uniq:
        print "##refname\trefstart\tctgname\tctgstart\tlength\tn\torient"
        for mem in idx.getmums(args.minlength):
            refstart=mem[2][0]
            ctgstart=mem[2][1]
            rnode=t[refstart].pop() #start position on match to node in graph
            cnode=t[ctgstart].pop()
            print "%s\t%s\t%s\t%s\t%s\t%s\t%s" % (rnode[2], refstart-rnode[0], cnode[2], ctgstart-cnode[0], mem[0], mem[1], 0)
    else:
        print "##refname\trefstart\tctgname\tctgstart\tlength\tn\tunique\torient"
        for mem in idx.getmems(args.minlength):
            refstart=mem[2][0]
            ctgstart=mem[2][1]
            rnode=t[refstart].pop() #start position on match to node in graph
            cnode=t[ctgstart].pop()
            print "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (rnode[2], refstart-rnode[0], cnode[2], ctgstart-cnode[0], mem[0], mem[1], mem[3], 0)
    
    if args.rc:

        logging.debug("Indexing reverse complement...\n")
        
        ### index reverse complement
        if args.sa64:
            idx=reveallib64.index(sa=args.sa2, lcp=args.lcp2) #enable preconstruction of second SA and LCP array
        else:
            idx=reveallib.index(sa=args.sa2, lcp=args.lcp2) #enable preconstruction of second SA and LCP array
        
        rcG=nx.DiGraph()
        t=IntervalTree()
        
        idx.addsample(reffile)
        if args.reference.endswith(".gfa"):
            read_gfa(args.reference,idx,t,rcG)
        else:
            rcG.graph['paths']=set([reffile])
            for name,seq in fasta_reader(args.reference):
                intv=idx.addsequence(seq)
                intv=Interval(intv[0],intv[1],name)
                t.add(intv)
                rcG.add_node(intv,offsets={reffile:0},aligned=0)
                refseq=seq
        
        idx.addsample(ctgfile)
        if args.contigs.endswith(".gfa"):
            read_gfa(args.contigs,idx,t,rcG,revcomp=True)
        else:
            rcG.graph['paths']=set([ctgfile])
            for name,seq in fasta_reader(args.contigs):
                intv=idx.addsequence(rc(seq))
                intv=Interval(intv[0],intv[1],name)
                t.add(intv)
                rcG.add_node(intv,offsets={ctgfile:0},aligned=0)
        
        idx.construct()
        
        if args.uniq:
            for mem in idx.getmums(args.minlength):
                refstart=mem[2][0]
                ctgstart=mem[2][1]
                rnode=t[refstart].pop() #start position on match to node in graph
                cnode=t[ctgstart].pop()
                l=cnode[1]-cnode[0]
                print "%s\t%s\t%s\t%s\t%s\t%s\t%s" % (rnode[2], refstart-rnode[0], cnode[2], l-((ctgstart-cnode[0])+mem[0]), mem[0], mem[1], 1)
        else:
            for mem in idx.getmems(args.minlength):
                refstart=mem[2][0]
                ctgstart=mem[2][1]
                rnode=t[refstart].pop() #start position on match to node in graph
                cnode=t[ctgstart].pop()
                l=cnode[1]-cnode[0]
                print "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % (rnode[2], refstart-rnode[0], cnode[2], l-((ctgstart-cnode[0])+mem[0]), mem[0], mem[1], mem[3], 1)



import logging
import traceback

from utils import *

from collections import defaultdict, deque
import os

import reveallib
import reveallib64
import schemes
import bubbles

def breaknode(node,pos,l):
    att=G.node[node]

    in_edges=[(e[0],e[2]) for e in G.in_edges(node,data=True)]
    out_edges=[(e[1],e[2]) for e in G.out_edges(node,data=True)]

    mn=Interval(pos,pos+l)
    other=set()
    
    if mn==node: #no breaking needed
        logging.log(1,"Node %s does not need to be broken."%str(node))
        t.remove(node)
        return node,other
    
    logging.log(1,"Breaking node: %s into: %s"%(str(node),str(mn)))

    allpaths=set()
    
    moffsets=dict()
    for s in att['offsets']:
        moffsets[s]=att['offsets'][s]+(pos-node.begin)
        allpaths.add(s)
    
    logging.log(1,"Offsets after break: %s"%str(moffsets))

    soffsets=dict()
    for s in att['offsets']:
        soffsets[s]=att['offsets'][s]+((pos+l)-node.begin)
    
    #if node is traversed via the other strand, add reverse edges
    negstrand=False
    negpaths=set()
    pospaths=set()
    
    if len(in_edges)==0 and len(out_edges)==0:
        pospaths=allpaths
    else:
        if len(in_edges)>0:
            for fro,d in in_edges:
                if d['oto']=='-':
                    negstrand=True
                    for p in d['paths']:
                        negpaths.add(p)
                else:
                    assert(d['oto']=='+')
                    for p in d['paths']:
                        pospaths.add(p)
        if len(out_edges)>0:
            for to,d in out_edges:
                if d['ofrom']=='-':
                    negstrand=True
                    for p in d['paths']:
                        negpaths.add(p)
                else:
                    assert(d['ofrom']=='+')
                    for p in d['paths']:
                        pospaths.add(p)

    if pospaths.intersection(negpaths)!=set():
        logging.error("Unable to properly separate paths through node: %s [%s,%s,%s,%s]"%(node, G.graph['id2path'], allpaths, pospaths, negpaths))

    assert(pospaths.intersection(negpaths)==set())

    G.add_node(mn,offsets=moffsets,aligned=0)#create merge node
    
    if (node[0]!=pos):
        pn=Interval(node[0],pos)
        logging.log(1,"Creating prefix node: %s"%str(pn))
        G.add_node(pn,offsets=att['offsets'],aligned=0)#create prefix node
        assert(not G.has_edge(pn,mn))
        assert(not G.has_edge(mn,pn))
        G.add_edge(pn,mn,paths=pospaths.copy(),ofrom='+',oto='+')
        assert(pospaths!=set())
        if negstrand:
            G.add_edge(mn,pn,paths=negpaths.copy(),ofrom='-',oto='-')
            assert(negpaths!=set())
        t.add(pn)
        other.add(pn)
    else:
        pn=mn

    if (node[1]!=pos+l):
        sn=Interval(pos+l,node[1])
        logging.log(1,"Creating suffix node: %s"%str(sn))
        G.add_node(sn,offsets=soffsets,aligned=0)#create suffix node
        assert(not G.has_edge(mn,sn))
        assert(not G.has_edge(sn,mn))
        G.add_edge(mn,sn,paths=pospaths.copy(),ofrom='+',oto='+')
        assert(pospaths!=set())
        if negstrand:
            G.add_edge(sn,mn,paths=negpaths.copy(),ofrom='-',oto='-')
            assert(negpaths!=set())
        t.add(sn)
        other.add(sn)
    else:
        sn=mn

    G.remove_node(node)                     #update Graph
    t.remove(node)                          #update intervaltree

    for fro,d in in_edges:
        if d['oto']=="+":
            G.add_edge(fro,pn,**d)
        else:
            G.add_edge(fro,sn,**d)

    for to,d in out_edges:
        if d['ofrom']=="+":
            G.add_edge(sn,to,**d)
        else:
            G.add_edge(pn,to,**d)
    
    logging.log(1,"Leading/Trailing node(s): %s"%str(other))
    logging.log(1,"Matching node: %s"%str(mn))

    return mn,other #return merge node

def mergenodes(G,mns):
    # logging.trace("Merging nodes %s"%str(mns))
    
    # global o
    ri=0
    # if reference!=None:
    #     for i,node in enumerate(mns):
    #         if reference in G.node[node]['offsets'].keys():
    #             refnode=node
    #             ri=i
    #             break
    #     else:
    #         refnode=mns[ri]
    # else:
    refnode=mns[ri]

    #merge the offset dictionaries
    newoffsets=dict()
    for node in mns:
        d=G.node[node]
        for sampleid in d['offsets']:
            if sampleid in newoffsets:
                logging.warn("WARNING: merging nodes that originate from the same sample: %s in %s."%(sample,str(newoffsets.keys())))
            #assert(sample not in newoffsets)
            newoffsets[sampleid]=d['offsets'][sampleid]
    
    G.node[refnode]['offsets']=newoffsets
    assert(len(G.node[refnode]['offsets'])==len(newoffsets))
    
    G.node[refnode]['aligned']=1
    
    tmp=mns.pop(ri)
    assert(tmp==refnode)

    for mn in mns: #leave the first node, merge the rest

        if type(G)==nx.MultiDiGraph:
            for e0,e1,k,d in G.in_edges(mn,keys=True,data=True):
                for _e0,_e1,_k,_d in G.in_edges(refnode,keys=True,data=True):
                    if type(_e0)==type(e0) and _e0==e0 and _d['oto']==d['oto'] and _d['ofrom']==d['ofrom']: #edge already exists, merge paths
                        for p in d['paths']:
                            G[_e0][_e1][_k]['paths'].add(p)
                        break
                else:
                    G.add_edge(e0,refnode,**d)

            for e0,e1,k,d in G.out_edges(mn,keys=True,data=True):
                for _e0,_e1,_k,_d in G.out_edges(refnode,keys=True,data=True):
                    if type(_e1)==type(e1) and _e1==e1 and _d['oto']==d['oto'] and _d['ofrom']==d['ofrom']: #edge already exists, merge paths
                        for p in d['paths']:
                            G[_e0][_e1][_k]['paths'].add(p)
                        break
                else:
                    G.add_edge(refnode,e1,**d)
        else:
            for e0,e1,d in G.in_edges(mn,data=True):
                if G.has_edge(e0,refnode):
                    for p in d['paths']:
                        G[e0][refnode]['paths'].add(p)
                else:
                    G.add_edge(e0,refnode,**d)
            for e0,e1,d in G.out_edges(mn,data=True):
                if G.has_edge(refnode,e1):
                    for p in d['paths']:
                        G[refnode][e1]['paths'].add(p)
                else:
                    G.add_edge(refnode,e1,**d)

        G.remove_node(mn)
    
    return refnode

def predecessorsfilter_iter(G,node):
    if type(G)==nx.MultiDiGraph:
        for pre in G.predecessors(node):
            for i in G[pre][node]:
                for p in G[pre][node][i]['paths']:
                    if not G.graph['id2path'][p].startswith("*"):
                        yield pre
                        break
    else:
        for pre in G.predecessors(node):
            for p in G[pre][node]['paths']:
                if not G.graph['id2path'][p].startswith("*"):
                    yield pre
                    break

def successorsfilter_iter(G,node):
    if type(G)==nx.MultiDiGraph:
        for suc in G.successors(node):
            for i in G[node][suc]:
                for p in G[node][suc][i]['paths']:
                    if not G.graph['id2path'][p].startswith("*"):
                        yield suc
                        break
    else:
        for suc in G.successors(node):
            for p in G[node][suc]['paths']:
                if not G.graph['id2path'][p].startswith("*"):
                    yield suc
                    break

def bfs(G, source, reverse=False, ignore=set()):
    if reverse:
        neighbors = predecessorsfilter_iter
    else:
        neighbors = successorsfilter_iter
    visited = set([source])
    queue = deque([(source, neighbors(G,source))])
    while queue:
        parent, children = queue[0]
        try:
            child = next(children)
            if child not in visited:
                visited.add(child)
                if 'aligned' not in G.node[child]:
                    assert(type(child)==str) #has to be start or end node
                    yield child,2
                elif (G.node[child]['aligned']==0):
                    queue.append((child, neighbors(G,child)))
                    yield child,0
                elif (G.node[child]['aligned']!=0 and child in ignore): #keep searching
                    queue.append((child, neighbors(G,child)))
                    yield child,0
                else:
                    yield child,1
        except StopIteration:
            queue.popleft()

def segmentgraph(node,nodes):
    trailing=set()
    leading=set()
    reverse_trailing=set()
    reverse_leading=set()
    nodes=set(nodes)
    
    #forward search
    endpoints=set()
    fwdmerge=set()
    for c,t in bfs(G,node):
        if t==0:
            trailing.add(c)
        elif t==1 or t==2:
            endpoints.add(c)
            if t==2:
                fwdmerge.add(c)
        else:
            logging.error("Node traversal failed, encountered: %s"%str((c,t)))
            sys.exit(1)
    
    #reverse search for each endpoint
    if len(endpoints)>1:
        for endpoint in endpoints:
            for c,t in bfs(G,endpoint,reverse=True,ignore=endpoints):
                if t==0:
                    reverse_trailing.add(c)
        trailing=trailing.intersection(reverse_trailing)
    
    #backward search
    endpoints=set()
    bwdmerge=set()
    for c,t in bfs(G,node,reverse=True):
        if t==0:
            leading.add(c)
        elif t==1 or t==2:
            endpoints.add(c)
            if t==2:
                bwdmerge.add(c)
        else:
            logging.error("Node traversal failed, encountered: %s"%str((c,t)))
            sys.exit(1)
    
    #reverse search for each endpoint
    if len(endpoints)>1:
        for endpoint in endpoints:
            for c,t in bfs(G,endpoint,ignore=endpoints):
                if t==0:
                    reverse_leading.add(c)
        leading=leading.intersection(reverse_leading)
    
    leading = set([(i.begin,i.end) for i in leading if isinstance(i,Interval)]).intersection(nodes) #TODO: remove "if isinstance(i,Interval)]"
    trailing = set([(i.begin,i.end) for i in trailing if isinstance(i,Interval)]).intersection(nodes)
    
    rest = nodes - (leading | trailing)
    
    return leading, trailing, rest

def graphalign(index,mum):
    try:
        logging.debug("In graphalign with %s"%str(mum))
        l,n,spd=mum
        nodes=index.nodes
        isize=index.n
        mns=[]
        topop=[]
        logging.debug("Nodes in subgraph:")
        for node in nodes:
            logging.debug("%s"%str(node))

        sp=[sp for gid,sp in spd]
        matching=set()
        for pos in sp:
            matching.add((pos,pos+l))
            # logging.debug("Lookup node for sp=%d"%pos)
            old=t[pos].pop()
            assert(old.end-old.begin>=l)
            mn,other=breaknode(old,pos,l)
            mns.append(mn)
            if isinstance(old,Interval):
                nodes.remove((old.begin,old.end))
            for node in other:
                if isinstance(node,Interval):
                    nodes.add((node.begin,node.end))
        
        mn=mergenodes(G,mns)
        msamples=set(G.node[Interval(mn[0],mn[1])]['offsets'].keys())
        # logging.trace("Merging samples: %s"%str(msamples))
        # logging.trace("Nodes before segmenting: %s"%nodes)

        intervals=segmentgraph(mn,nodes)
        leading,trailing,rest=intervals

        # logging.trace("Leading nodes after segmenting: %s"%leading)
        # logging.trace("Trailing nodes after segmenting: %s"%trailing)
        # logging.trace("Parallel nodes after segmenting: %s"%rest)

        logging.debug("Merged interval: %s"%str(mn))
        logging.debug("Number of leading intervals: %d"%len(leading))
        logging.debug("Number of trailing intervals: %d"%len(trailing))
        logging.debug("Number of parallel intervals: %d"%len(rest))
        # logging.trace("Number of nodes in the entire graph: %d"%G.number_of_nodes())
        newleftnode=mn
        newrightnode=mn

        for intv in leading:
            if not set(G.node[Interval(intv[0],intv[1])]['offsets'].keys()).issubset(msamples): #no clean dissection of all paths on the left
                newrightnode=index.rightnode
                break
        
        for intv in trailing:
            if not set(G.node[Interval(intv[0],intv[1])]['offsets'].keys()).issubset(msamples): #no clean dissection of all paths on the right
                newleftnode=index.leftnode
                break

        return leading,trailing,matching,rest,mn,newleftnode,newrightnode

    except Exception:
        print traceback.format_exc()
        raise Exception
        return

#TODO: rewrite so this uses bubble definition code
def prune_nodes(G,T=""):
    converged=False
    while not(converged):
        converged=True
        for node,data in G.nodes(data=True):
            if node not in G:
                continue

            for run in [0,1]:
                # if type(node)==str or data['aligned']!=0:
                if run==0:
                    if type(G)==nx.MultiDiGraph:
                        neis=[n2 for n1,n2,k,d in G.out_edges(node,keys=True,data=True) if d['ofrom']=='+' and d['oto']=='+']
                    else:
                        neis=[n2 for n1,n2,d in G.out_edges(node,data=True) if d['ofrom']=='+' and d['oto']=='+']
                else:
                    if type(G)==nx.MultiDiGraph:
                        neis=[n1 for n1,n2,k,d in G.in_edges(node,keys=True,data=True) if d['ofrom']=='+' and d['oto']=='+']
                    else:
                        neis=[n1 for n1,n2,d in G.in_edges(node,data=True) if d['ofrom']=='+' and d['oto']=='+']
                
                seqs={}
                for nei in neis:
                    if 'seq' not in G.node[nei]:
                        if not isinstance(nei,Interval):
                            continue
                        seq=T[nei.begin:nei.end]
                    else:
                        seq=G.node[nei]['seq']
                    if seq in seqs:
                        seqs[seq].append(nei)
                    else:
                        seqs[seq]=[nei]

                for key in seqs:
                    group=seqs[key]
                    if len(group)>1:
                        merge=True
                        for v in group:
                            if run==0:
                                if type(G)==nx.MultiDiGraph:
                                    if len([n1 for n1,n2,k,d in G.in_edges(v,keys=True,data=True) if d['ofrom']=='+' and d['oto']=='+'])>1:
                                        merge=False
                                        break
                                else:
                                    if len([n1 for n1,n2,d in G.in_edges(v,data=True) if d['ofrom']=='+' and d['oto']=='+'])>1:
                                        merge=False
                                        break
                            else:
                                if type(G)==nx.MultiDiGraph:
                                    if len( [n2 for n1,n2,k,d in G.out_edges(v,keys=True,data=True) if d['ofrom']=='+' and d['oto']=='+'] )>1:
                                        merge=False
                                        break
                                else:
                                    if len( [n2 for n1,n2,d in G.out_edges(v,data=True) if d['ofrom']=='+' and d['oto']=='+'] )>1:
                                        merge=False
                                        break
                        if merge:
                            mergenodes(G,group)
                            converged=False

def align_cmd(args):
    G,idx=align_genomes(args)
    
    if args.output==None:
        pref=[]
        for f in args.inputfiles:
            bn=os.path.basename(f)
            if '.' in bn:
                pref.append(bn[:bn.find('.')])
            else:
                pref.append(bn)
        args.output="_".join(pref)
        args.output+=".gfa"
    
    logging.info("Merging nodes...")
    T=idx.T

    if len(G.graph['paths'])>2:
        prune_nodes(G,T=T)

    seq2node(G,T,remap=False)
    
    logging.info("Done.")
    
    alignedbases=0
    alignednodes=0
    
    totnodes=G.number_of_nodes()

    #TODO: start, report identity per sample that was aligned, this does not make sense...
    if idx.nsamples>2: #was multi-alignment
        totbases=idx.n-T.count('$')-T.count('N')
        for node,data in G.nodes(data=True):
            if 'aligned' in data and data['aligned']!=0 and type(node)!=str:
                alignedbases+=(node.end-node.begin)*len([k for k in data['offsets'] if not G.graph['id2path'][k].startswith("*")])
                alignednodes+=1
    else: #assume seq to graph
        totbases=idx.n-T.count('$')-T.count('N') # min([(idx.n-1)-(idx.nsep[0]+1),idx.nsep[0]])
        for node,data in G.nodes(data=True):
            if 'aligned' in data and data['aligned']!=0 and type(node)!=str:
                l=node.end-node.begin
                alignedbases+=(l*2)
                alignednodes+=1
    
    logging.info("%s (%.2f%% identity, %d bases out of %d aligned, %d nodes out of %d aligned)."%("-".join([os.path.basename(f) for f in args.inputfiles]), (alignedbases/(float(totbases)))*100,alignedbases,totbases,alignednodes,totnodes))
    #TODO: end

    logging.info("Writing graph...")
    if args.gml:
        graph=write_gml(G,T, hwm=args.hwm, outputfile=args.output, partition=False)
    else:
        write_gfa(G,T, outputfile=args.output)
        graph=args.output

    logging.info("Done.")
    logging.info("Graph written to: %s"%graph)
    
    if args.mumplot:
        if len(G.graph['paths'])==2:
            plotgraph(G,G.graph['paths'][0],G.graph['paths'][1],interactive=args.interactive)
        else:
            logging.info("Unable to make plot for graphs with more than 2 paths.")
    
    #plotgraph(G,G.graph['paths'][0],G.graph['paths'][1],interactive=args.interactive)

def align_genomes(args):
    logging.info("Loading input...")
    #global variables to simplify callbacks from c extension
    global t,G
    
    # global reference
    # reference=args.reference
    
    t=IntervalTree()

    if args.sa64:
        idx=reveallib64.index(sa=args.sa, lcp=args.lcp, cache=args.cache)
    else:
        idx=reveallib.index(sa=args.sa, lcp=args.lcp, cache=args.cache)
    
    #G=nx.DiGraph()
    G=nx.MultiDiGraph()

    o=0
    schemes.args=args
    
    graph=False
    
    for i,sample in enumerate(args.inputfiles):
        
        if sample.endswith(".gfa"):
            idx.addsample(os.path.basename(sample))
            graph=True

            logging.info("Reading graph: %s ..." % sample)
            if i==0:
                read_gfa(sample,idx,t,G,minsamples=args.minsamples,
                                        maxsamples=args.maxsamples,
                                        targetsample=args.targetsample,
                                        remap=True)
            else:
                read_gfa(sample,idx,t,G,remap=True)

        else: #consider it to be a fasta file
            read_fasta(sample,idx,t,G,contigs=args.contigs,toupper=args.toupper)
    
    logging.debug("Graph contains the following paths: %s"%G.graph['paths'])

    logging.debug("Index contains the following samples: %s"%idx.samples)

    if len(idx.samples)<=1:
        logging.fatal("Specify at least 2 targets to construct alignment. In case of multi-fasta, consider the --nocontigs flag.")
        sys.exit(1)

    if not nx.is_directed_acyclic_graph(G):
        logging.info("*** Input is not a DAG! ...")

    for n1,n2,data in G.edges(data=True):
        assert('paths' in data)

    schemes.ts=t
    schemes.G=G
    
    logging.info("Constructing index...")
    idx.construct()
    
    logging.info("Done.")
    
    if len(args.inputfiles)==2 and not graph:
        logging.info("Constructing pairwise-alignment...")
        idx.align(schemes.graphmumpicker,graphalign,threads=args.threads,wpen=args.wpen,wscore=args.wscore,minl=args.minlength,minn=args.minn)
    else:
        logging.info("Constructing graph-based multi-alignment...")
        idx.align(schemes.graphmumpicker,graphalign,threads=args.threads,wpen=args.wpen,wscore=args.wscore,minl=args.minlength,minn=args.minn)
    
    # from multiprocessing import Process
    # from Queue import Queue
    # main=idx #make sure we keep the main ref count, since it has the reference to T
    # q=Queue()
    # q.put(idx)
    # while not q.empty():
    #     idx=q.get()
    #     if len(args.inputfiles)>2:
    #         multimums=idx.getmultimums(minlength=args.minlength, minn=args.minn)
    #     else:
    #         multimums=idx.mums(args.minlength)
    #     if len(multimums)==0:
    #         continue
    #     ret=schemes.graphmumpicker(multimums,idx)
    #     if ret==None:
    #         continue
    #     else:
    #         splitmum,skipleft,skipright=ret
    #     ret=graphalign(idx,splitmum)
    #     if ret==None:
    #         continue
    #     else:
    #         leading,trailing,matching,rest,merged,newleftnode,newrightnode=ret
    #     ilead,itrail,ipar=idx.splitindex(leading,trailing,matching,rest,merged,newleftnode,newrightnode,skipleft,skipright)
    #     if ilead!=None and ilead.n>1:
    #         q.put(ilead)
    #     if itrail!=None and itrail.n>1:
    #         q.put(itrail)
    #     if ipar!=None and ipar.n>1:
    #         q.put(ipar)

    return G,idx


#seq should be a list of objects that can be (multi-) aligned by reveal:
#   - tuple of the form (name,seq)
def align(aobjs,ref=None,minlength=20,minn=2,seedsize=None,threads=0,targetsample=None,maxsamples=None,\
                maxmums=10000,wpen=1,wscore=1,sa64=False,pcutoff=1e-8,gcmodel="sumofpairs",maxsize=None,\
                trim=True):
    
    kwargs = dict(locals()) #hack the kwargs into a dict so we can pass it to schemes as if it were the argparsed args object
    class dict2class(object):
        def __init__(self, d):
            self.__dict__ = d
    args=dict2class(kwargs)
    schemes.args=args
    
    #global variables to simplify callbacks from c extension
    global t,G

    t=IntervalTree()

    if sa64:
        idx=reveallib64.index()
    else:
        idx=reveallib.index()
    
    G=nx.DiGraph()

    G.graph['paths']=[]
    G.graph['path2id']=dict()
    G.graph['id2path']=dict()
    G.graph['id2end']=dict()
    o=0

    graph=False
    
    startnode=uuid.uuid4().hex
    G.add_node(startnode)
    endnode=uuid.uuid4().hex
    G.add_node(endnode)

    for aobj in aobjs:
        if isinstance(aobj,tuple):
            name,seq=aobj
            idx.addsample(name)
            intv=idx.addsequence(seq.upper())
            if intv[1]-intv[0]>0:
                Intv=Interval(intv[0],intv[1])
                t.add(Intv)
                sid=len(G.graph['paths'])
                G.graph['path2id'][name]=len(G.graph['paths'])
                G.graph['id2path'][sid]=name
                G.graph['id2end'][sid]=len(seq)
                
                # G.node[endnode]['offsets'][sid]=len(seq)
                # G.node[startnode]['offsets'][sid]=0

                G.graph['paths'].append(name)
                G.add_node(Intv,offsets={sid:0},aligned=0)
                G.add_edge(startnode,Intv,paths={sid},ofrom='+',oto='+')
                G.add_edge(Intv,endnode,paths={sid},ofrom='+',oto='+')

        # elif isinstance(aobj,str):
        #     if not os.path.isfile(aobj):
        #         logging.fatal("Not a file, expecting fasta or gfa file.")
        #         return
        #     idx.addsample(os.path.basename(aobj))
        #     if aobj.endswith(".gfa"):
        #         read_gfa(aobj,idx,t,G,targetsample=targetsample,maxsamples=maxsamples)
        #         graph=True
        #     else: #assume a file in fastaformat
        #         for name,seq in fasta_reader(sample):
        #             intv=idx.addsequence(seq.upper())
        #             if intv[1]-intv[0]>0:
        #                 Intv=Interval(intv[0],intv[1])
        #                 t.add(Intv)
        #                 sid=len(G.graph['paths'])
        #                 G.graph['path2id'][name]=len(G.graph['paths'])
        #                 G.graph['id2path'][sid]=name
        #                 G.graph['id2end'][sid]=len(seq)
        #                 G.graph['paths'].append(name)
        #                 G.add_node(Intv,offsets={sid:0},aligned=0)
        #                 G.add_edge(startnode,Intv,paths={sid})
        #                 G.add_edge(endnode,Intv,paths={sid})
    
    if not nx.is_directed_acyclic_graph(G):
        logging.error("*** Input is not a DAG! Not supported.")
        return
    
    schemes.ts=t
    schemes.G=G
    
    idx.construct()
    
    idx.align(schemes.graphmumpicker,graphalign,threads=threads,wpen=wpen,wscore=wscore,minl=minlength,minn=minn)

    prune_nodes(G,T=idx.T)

    G.remove_node(startnode)
    G.remove_node(endnode)

    return G,idx


import networkx as nx
import utils
import sys
import logging
import re

def extract_cmd(args):
    if not args.graph[0].endswith(".gfa"):
        logging.fatal("Invalid gfa file.")
        return
    width=args.width

    if args.nocycles:
        G=nx.DiGraph()
    else:
        G=nx.MultiDiGraph()

    utils.read_gfa(args.graph[0], None, None, G, remap=False)

    if args.all:
        args.input=sorted(G.graph['paths'])
    
    try:
        i=0
        for ins in args.input:
            if args.type=="pathname":
                seq=extract(G,ins)
            elif args.type=="path":
                seq=extract_path(G,ins.split(","))
            else:
                logging.fatal("Unknown input type")
                sys.exit(1)
            
            sys.stdout.write(">"+ins+"\n")
            f=0
            for i in xrange(width,len(seq),width):
                sys.stdout.write(seq[f:i]+'\n')
                f=i
            sys.stdout.write(seq[f:]+'\n')
            i+=1
    except IOError:
        try:
            sys.stdout.close()
        except IOError:
            pass
        try:
            sys.stderr.close()
        except IOError:
            pass

#TODO: contribute patch back to networkx
def dag_longest_path_custom(G, weight='weight', default_weight=1):
    if not G:
        return []
    dist = {}  # stores {v : (length, u)}
    for v in nx.topological_sort(G):

        if type(G)==nx.MultiDiGraph:
            us = [(dist[u][0] + max([data[k].get(weight, default_weight) for k in data]), u)
                  for u, data in G.pred[v].items()]
        else:
            us = [(dist[u][0] + data.get(weight, default_weight), u)
                  for u, data in G.pred[v].items()]

        # Use the best predecessor if there is one and its distance is
        # non-negative, otherwise terminate.
        maxu = max(us, key=lambda x: x[0]) if us else (0, v)
        dist[v] = maxu if maxu[0] >= 0 else (0, v)
    u = None
    v = max(dist, key=lambda x: dist[x][0])
    path = []
    while u != v:
        path.append(v)
        u = v
        v = dist[v][1]
    path.reverse()
    return path


def extract(G,sample):
    logging.info("Extracting path: %s from graph (%s) of size: (%d,%d)"%(sample,type(G),G.number_of_nodes(),G.number_of_edges()))
    
    if sample == "_longest_":
        #shortcut to extract the "longest" path in terms of sequence

        if type(G)==nx.MultiDiGraph:
            sv=utils.MultiGraphToDiGraph(G)
            for v,t,k in G.edges:
                G[v][t][k]['weight']=len(G.node[t]['seq'])-G.node[t]['seq'].count("N") if 'seq' in G.node[t] else 0
        else:
            for v,t in G.edges:
                G[v][t]['weight']=len(G.node[t]['seq'])-G.node[t]['seq'].count("N") if 'seq' in G.node[t] else 0

        # p=[]
        seq=""
        # e=None
        # weights=[0]
        for n in dag_longest_path_custom(G, weight='weight'):
            # p.append(n)
            # if e!=None:
            #     if 0 in G[e][n]:
            #         weights.append(G[e][n][0]['weight'])
            #     else:
            #         weights.append(G[e][n]['weight'])
            seq+=G.node[n]['seq']
            # e=n

        # with open("path.txt",'w') as f:
        #     f.write("total length: %d\n"%sum(weights))
        #     for n,w in zip(p,weights):
        #         f.write("%s-%d\n"%(n,w))

        return seq
        
    elif sample not in G.graph['path2id']:
        logging.fatal("Unknown path: %s, graph contains: %s"%(sample, G.graph['path2id'].keys()))
        sys.exit(1)

    else:
        sid=G.graph['path2id'][sample]
        
        sg=[]
        for n1,n2,d in G.edges(data=True):
            if sid in d['paths']:
                sg.append((n1,n2,d))
        
        if len(sg)>0:
            #G can be a MultiDiGraph, but subgraph should be single edge!
            subgraph=nx.DiGraph(sg)
            seq=""
            path=list(nx.topological_sort(subgraph))

            if type(G)==nx.MultiDiGraph:
                inito=G[path[0]][path[1]][0]['ofrom']
            else:
                inito=G[path[0]][path[1]]['ofrom']

            pnode=None

            for node in path:
                offset=0
                if pnode==None:
                    o=inito
                else:
                    o=subgraph[pnode][node]['oto']
                    if 'cigar' in subgraph[pnode][node] and subgraph[pnode][node]['cigar']!='0M':
                        cigar=subgraph[pnode][node]['cigar']
                        a=re.findall(r'(\d+)(\w)', cigar)
                        for l,t in a: #determine offset within the segment to allow for overlapping segments
                            if t=='M' or t=='I' or t=='S' or t=='P': #source of the edge (pnode) is considered the reference
                                offset+=int(l)
                    
                if o=="+":
                    s=G.node[node]['seq']
                else:
                    s=utils.rc(G.node[node]['seq'])

                assert(len(s)>=offset)

                seq+=s[offset:]
                pnode=node

        else: #has to be a single node
            seq=""
            for n in G:
                if sid in G.node[n]['offsets']:
                    seq=G.node[n]['seq']
                    break

        return seq

def extract_path(G,path,type="str"):
    logging.debug("Extracting path of length: %d"%len(path))

    seq=""
    for n in path:
        nid,o=int(n[:-1]),n[-1:]
        assert(o=='+' or o=='-')

        if o=="+":
            seq+=G.node[nid]['seq']
        else:
            seq+=utils.rc(G.node[nid]['seq'])

    # return seq


import networkx as nx
from utils import *
import sys
from bubbles import bubbles,Bubble,Variant

def stats_cmd(args):
    if len(args.gfa)!=1:
        logging.fatal("Specify 1 gfa file.")
        return
    
    if not args.gfa[0].endswith('.gfa'):
        logging.fatal("Use .gfa as extension of the gfa file.")
        return
    
    stats(args.gfa[0])

def stats(gfafile):
    stats=dict()
    
    G=nx.MultiDiGraph()
    read_gfa(gfafile,None,"",G)

    struct=MultiGraphToDiGraph(G)

    samples=G.graph['paths']
    nsamples=len(samples)
    
    stats["Graph"]=os.path.basename(gfafile)
    stats["Number of samples"]=nsamples
    for i,sample in enumerate(samples):
        stats["Sample %d"%i]=sample

    stats["Number of rearrangement edges"]=len(struct)

    stats["Number of connected components"]=0
    
    stats["Count A"]=0
    stats["Count C"]=0
    stats["Count G"]=0
    stats["Count T"]=0
    stats["Count N"]=0
    for node,data in G.nodes(data=True):
        stats["Count A"]+=data['seq'].count('A')
        stats["Count C"]+=data['seq'].count('C')
        stats["Count G"]+=data['seq'].count('G')
        stats["Count T"]+=data['seq'].count('T')
        stats["Count N"]+=data['seq'].count('N')

    seqperngenomes=dict()
    
    i=1
    for sample in G.graph['paths']:
        seqperngenomes[i]=0
        i+=1
    
    for node,data in G.nodes(data=True):
        seqperngenomes[len([o for o in data['offsets'] if not G.graph['id2path'][o].startswith("*")])]+=len(data['seq'])

    for n in seqperngenomes:
        stats["Sequence observed in %d genomes"%n]=seqperngenomes[n]

    #for each connected component
    for sgi,sub in enumerate(nx.weakly_connected_component_subgraphs(G)):
        stats["Number of connected components"]+=1
        sg=G.subgraph(sub)
        
        #determine samples in subgraph
        nsgsamples=1
        sgsamples=set()
        sgsampleids=set()
        for node,data in sg.nodes(data=True):
            if len(data['offsets'])>nsgsamples:
                nsgsamples=len([o for o in data['offsets'] if not sg.graph['id2path'][o].startswith("*")])

            for sid in data['offsets']:
                if not sg.graph['id2path'][sid].startswith("*"):
                    sgsamples.add(G.graph['id2path'][sid])
                    sgsampleids.add(sid)
        
        stats["Composition of component %d"%sgi]=",".join(sgsamples)

        #count bubble stats
        complexbubbles=0
        simplebubbles=0
        snpcount=0
        indelcount=0
        multicount=0
        regioncount=0
        sgapcount=0
        cgapcount = 0
        unknowncount=0
        
        for bubble in bubbles(sg):
            if bubble.issimple():
                simplebubbles+=1
            else:
                complexbubbles+=1
            
            v=Variant(bubble)
            
            if v.vtype=='snp':
                snpcount+=1
            elif v.vtype=='indel':
                indelcount+=1
            elif v.vtype=='multi-allelic':
                multicount+=1
            elif v.vtype=='region':
                regioncount+=1
            else:
                unknowncount+=1

            if v.gap=='simplegap':
                sgapcount+= 1
            elif v.gap=='complexgap':
                cgapcount+=1
        
        stats["Number of bubbles in component %d (total)"%sgi]=complexbubbles+simplebubbles
        stats["Number of bubbles in component %d (simple)"%sgi]=simplebubbles
        stats["Number of bubbles in component %d (complex)"%sgi]=complexbubbles
        stats["Number of variants in component %d (snps)"%sgi]=snpcount
        stats["Number of variants in component %d (indels)"%sgi]=indelcount
        stats["Number of variants in component %d (multi-allelic)"%sgi]=multicount
        stats["Number of variants in component %d (complex)"%sgi]=unknowncount
        stats["Number of variants in component %d (regions)"%sgi]=regioncount
        stats["Number of variants in component %d (simple gaps)"%sgi]=sgapcount
        stats["Number of variants in component %d (complex gaps)"%sgi]=cgapcount


        #chain stats
        chain=[]
        chainweight=0
        chainpenalty=0
        chainlength=0
        chainlengthbp=0
        
        for node,data in sg.nodes(data=True):
            if type(node)==str: #skip start and end nodes
                continue
            offsets=data['offsets']
            l=len(data['seq'])
            if set(offsets.keys())==sgsampleids:
                coords=tuple([offsets[k] for k in sorted(offsets.keys())])
                chain.append((coords,l))
                chainweight+=l*((len(offsets)*(len(offsets)-1))/2) #sumofpairs score!
                chainlengthbp+=l
                chainlength+=1
        
        if len(chain)>0:
            chain.sort(key=lambda l: l[0])
            
            ppoint=tuple([c+chain[0][1] for c in chain[0][0]])
            for point,length in chain[1:]:
                for i in range(len(point)):
                    assert(point[i]>=ppoint[i])
                p=gapcost(ppoint,point) #sumofpairs penalty!
                ppoint=tuple([c+length for c in point])
                chainpenalty+=p
        
        stats["Chain length in component %d"%sgi]=chainlength
        stats["Chain length basepairs in component %d"%sgi]=chainlengthbp
        stats["Chain weight (sum-of-pairs) in component %d"%sgi]=chainweight
        stats["Chain penalty (sum-of-pairs) in component %d"%sgi]=chainpenalty
        stats["Chain score in component %d"%sgi]=chainweight-chainpenalty

    remove=[]
    for node in G.nodes():
        if type(node)==str:
            remove.append(node)
    G.remove_nodes_from(remove)

    stats["Number of nodes"]=G.number_of_nodes()
    stats["Number of edges"]=G.number_of_edges()

    for label in sorted(stats.keys()):
        sys.stdout.write("%s:\t%s\n"%(label.ljust(50),str(stats[label]).rjust(50)))

